#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2020, Google Inc.
#
# Author: Paul Elder <paul.elder@ideasonboard.com>
#
# Generates libcamera files from a mojom.Module.

import argparse
import datetime
import os
import re

import mojom.fileutil as fileutil
import mojom.generate.generator as generator
import mojom.generate.module as mojom
from mojom.generate.template_expander import UseJinja


GENERATOR_PREFIX = 'libcamera'

_kind_to_cpp_type = {
    mojom.BOOL:   'bool',
    mojom.INT8:   'int8_t',
    mojom.UINT8:  'uint8_t',
    mojom.INT16:  'int16_t',
    mojom.UINT16: 'uint16_t',
    mojom.INT32:  'int32_t',
    mojom.UINT32: 'uint32_t',
    mojom.FLOAT:  'float',
    mojom.INT64:  'int64_t',
    mojom.UINT64: 'uint64_t',
    mojom.DOUBLE: 'double',
}

_bit_widths = {
    mojom.BOOL:   '8',
    mojom.INT8:   '8',
    mojom.UINT8:  '8',
    mojom.INT16:  '16',
    mojom.UINT16: '16',
    mojom.INT32:  '32',
    mojom.UINT32: '32',
    mojom.FLOAT:  '32',
    mojom.INT64:  '64',
    mojom.UINT64: '64',
    mojom.DOUBLE: '64',
}

def ModuleName(path):
    return path.split('/')[-1].split('.')[0]

def ModuleClassName(module):
    return re.sub(r'^IPA(.*)Interface$', lambda match: match.group(1),
                  module.interfaces[0].mojom_name)

def Capitalize(name):
    return name[0].upper() + name[1:]

def ConstantStyle(name):
    return generator.ToUpperSnakeCase(name)

def Choose(cond, t, f):
    return t if cond else f

def CommaSep(l):
    return ', '.join([m for m in l])

def ParamsCommaSep(l):
    return ', '.join([m.mojom_name for m in l])

def GetDefaultValue(element):
    if element.default is not None:
        return element.default
    if type(element.kind) == mojom.ValueKind:
        return '0'
    if IsFlags(element):
        return ''
    if mojom.IsEnumKind(element.kind):
        return f'static_cast<{element.kind.mojom_name}>(0)'
    if isinstance(element.kind, mojom.Struct) and \
       element.kind.mojom_name == 'SharedFD':
        return '-1'
    return ''

def HasDefaultValue(element):
    return GetDefaultValue(element) != ''

def HasDefaultFields(element):
    return True in [HasDefaultValue(x) for x in element.fields]

def GetAllTypes(element):
    if mojom.IsArrayKind(element):
        return GetAllTypes(element.kind)
    if mojom.IsMapKind(element):
        return GetAllTypes(element.key_kind) + GetAllTypes(element.value_kind)
    if isinstance(element, mojom.Parameter):
        return GetAllTypes(element.kind)
    if mojom.IsEnumKind(element):
        return [element.mojom_name]
    if not mojom.IsStructKind(element):
        return [element.spec]
    if len(element.fields) == 0:
        return [element.mojom_name]
    ret = [GetAllTypes(x.kind) for x in element.fields]
    ret = [x for sublist in ret for x in sublist]
    return list(set(ret))

def GetAllAttrs(element):
    if mojom.IsArrayKind(element):
        return GetAllAttrs(element.kind)
    if mojom.IsMapKind(element):
        return {**GetAllAttrs(element.key_kind), **GetAllAttrs(element.value_kind)}
    if isinstance(element, mojom.Parameter):
        return GetAllAttrs(element.kind)
    if mojom.IsEnumKind(element):
        return element.attributes if element.attributes is not None else {}
    if mojom.IsStructKind(element) and len(element.fields) == 0:
        return element.attributes if element.attributes is not None else {}
    if not mojom.IsStructKind(element):
        if hasattr(element, 'attributes'):
            return element.attributes or {}
        return {}
    attrs = [(x.attributes) for x in element.fields]
    ret = {}
    for d in attrs:
        ret.update(d or {})
    if hasattr(element, 'attributes'):
        ret.update(element.attributes or {})
    return ret

def NeedsControlSerializer(element):
    types = GetAllTypes(element)
    for type in ['ControlList', 'ControlInfoMap']:
        if f'x:{type}' in types:
            raise Exception(f'Unknown type "{type}" in {element.mojom_name}, did you mean "libcamera.{type}"?')
    return "ControlList" in types or "ControlInfoMap" in types

def HasFd(element):
    attrs = GetAllAttrs(element)
    if isinstance(element, mojom.Kind):
        types = GetAllTypes(element)
    else:
        types = GetAllTypes(element.kind)
    return "SharedFD" in types or (attrs is not None and "hasFd" in attrs)

def WithDefaultValues(element):
    return [x for x in element if HasDefaultValue(x)]

def WithFds(element):
    return [x for x in element if HasFd(x)]

def MethodParamInputs(method):
    return method.parameters

def MethodParamOutputs(method):
    if method.response_parameters is None:
        return []

    if MethodReturnValue(method) == 'void':
        return method.response_parameters

    if len(method.response_parameters) <= 1:
        return []

    return method.response_parameters[1:]

def MethodParamsHaveFd(parameters):
    return len([x for x in parameters if HasFd(x)]) > 0

def MethodInputHasFd(method):
    return MethodParamsHaveFd(method.parameters)

def MethodOutputHasFd(method):
    return MethodParamsHaveFd(MethodParamOutputs(method))

def MethodParamNames(method):
    params = []
    for param in method.parameters:
        params.append(param.mojom_name)
    for param in MethodParamOutputs(method):
        params.append(param.mojom_name)
    return params

def MethodParameters(method):
    params = []
    for param in method.parameters:
        params.append('const %s %s%s' % (GetNameForElement(param),
                                         '' if IsPod(param) or IsEnum(param) else '&',
                                         param.mojom_name))
    for param in MethodParamOutputs(method):
        params.append(f'{GetNameForElement(param)} *{param.mojom_name}')
    return params

def MethodReturnValue(method):
    if method.response_parameters is None or len(method.response_parameters) == 0:
        return 'void'
    first_output = method.response_parameters[0]
    if ((len(method.response_parameters) == 1 and IsPod(first_output)) or
        first_output.kind == mojom.INT32):
        return GetNameForElement(first_output)
    return 'void'

def IsAsync(method):
    # Events are always async
    if re.match("^IPA.*EventInterface$", method.interface.mojom_name):
        return True
    elif re.match("^IPA.*Interface$", method.interface.mojom_name):
        if method.attributes is None:
            return False
        elif 'async' in method.attributes and method.attributes['async']:
            return True
    return False

def IsArray(element):
    return mojom.IsArrayKind(element.kind)

def IsControls(element):
    return mojom.IsStructKind(element.kind) and (element.kind.mojom_name == "ControlList" or
                                                 element.kind.mojom_name == "ControlInfoMap")

def IsEnum(element):
    return mojom.IsEnumKind(element.kind)


# Only works the enum definition, not types
def IsScoped(element):
    attributes = getattr(element, 'attributes', None)
    if not attributes:
        return False
    return 'scopedEnum' in attributes


def IsEnumScoped(element):
    if not IsEnum(element):
        return False
    return IsScoped(element.kind)

def IsFd(element):
    return mojom.IsStructKind(element.kind) and element.kind.mojom_name == "SharedFD"


def IsFlags(element):
    attributes = getattr(element, 'attributes', None)
    if not attributes:
        return False
    return 'flags' in attributes

def IsMap(element):
    return mojom.IsMapKind(element.kind)

def IsPlainStruct(element):
    return mojom.IsStructKind(element.kind) and not IsControls(element) and not IsFd(element)

def IsPod(element):
    return element.kind in _kind_to_cpp_type

def IsStr(element):
    return element.kind.spec == 's'

def BitWidth(element):
    if element.kind in _bit_widths:
        return _bit_widths[element.kind]
    if mojom.IsEnumKind(element.kind):
        return '32'
    return ''

def ByteWidthFromCppType(t):
    key = None
    for mojo_type, cpp_type in _kind_to_cpp_type.items():
        if t == cpp_type:
            key = mojo_type
    if key is None:
        raise Exception('invalid type')
    return str(int(_bit_widths[key]) // 8)

# Get the type name for a given element
def GetNameForElement(element):
    # Flags
    if IsFlags(element):
        return f'Flags<{GetFullNameForElement(element.kind)}>'
    # structs
    if (mojom.IsEnumKind(element) or
        mojom.IsInterfaceKind(element) or
        mojom.IsStructKind(element)):
        return element.mojom_name
    # vectors
    if (mojom.IsArrayKind(element)):
        elem_name = GetFullNameForElement(element.kind)
        return f'std::vector<{elem_name}>'
    # maps
    if (mojom.IsMapKind(element)):
        key_name = GetFullNameForElement(element.key_kind)
        value_name = GetFullNameForElement(element.value_kind)
        return f'std::map<{key_name}, {value_name}>'
    # struct fields and function parameters
    if isinstance(element, (mojom.Field, mojom.Method, mojom.Parameter)):
        # maps and vectors
        if (mojom.IsArrayKind(element.kind) or mojom.IsMapKind(element.kind)):
            return GetNameForElement(element.kind)
        # strings
        if (mojom.IsReferenceKind(element.kind) and element.kind.spec == 's'):
            return 'std::string'
        # PODs
        if element.kind in _kind_to_cpp_type:
            return _kind_to_cpp_type[element.kind]
        # structs and enums
        return element.kind.mojom_name
    # PODs that are members of vectors/maps
    if (hasattr(element, '__hash__') and element in _kind_to_cpp_type):
        return _kind_to_cpp_type[element]
    if (hasattr(element, 'spec')):
        # strings that are members of vectors/maps
        if (element.spec == 's'):
            return 'std::string'
        # structs that aren't defined in mojom that are members of vectors/maps
        if (element.spec[0] == 'x'):
            return element.spec.replace('x:', '').replace('.', '::')
    if (mojom.IsInterfaceRequestKind(element) or
        mojom.IsAssociatedKind(element) or
        mojom.IsPendingRemoteKind(element) or
        mojom.IsPendingReceiverKind(element) or
        mojom.IsUnionKind(element)):
        raise Exception('Unsupported element: %s' % element)
    raise Exception('Unexpected element: %s' % element)

def GetFullNameForElement(element):
    name = GetNameForElement(element)
    namespace_str = ''
    if (mojom.IsStructKind(element) or mojom.IsEnumKind(element)):
        namespace_str = element.module.mojom_namespace.replace('.', '::')
    elif (hasattr(element, 'kind') and
          (mojom.IsStructKind(element.kind) or mojom.IsEnumKind(element.kind))):
        namespace_str = element.kind.module.mojom_namespace.replace('.', '::')

    if namespace_str == '':
        return name

    if IsFlags(element):
        return GetNameForElement(element)

    return f'{namespace_str}::{name}'

def ValidateZeroLength(l, s, cap=True):
    if l is None:
        return
    if len(l) > 0:
        raise Exception(f'{s.capitalize() if cap else s} should be empty')

def ValidateSingleLength(l, s, cap=True):
    if len(l) > 1:
        raise Exception(f'Only one {s} allowed')
    if len(l) < 1:
        raise Exception(f'{s.capitalize() if cap else s} is required')

def GetMainInterface(interfaces):
    intf = [x for x in interfaces
            if re.match("^IPA.*Interface", x.mojom_name) and
               not re.match("^IPA.*EventInterface", x.mojom_name)]
    ValidateSingleLength(intf, 'main interface')
    return None if len(intf) == 0 else intf[0]

def GetEventInterface(interfaces):
    event = [x for x in interfaces if re.match("^IPA.*EventInterface", x.mojom_name)]
    ValidateSingleLength(event, 'event interface')
    return None if len(event) == 0 else event[0]

def ValidateNamespace(namespace):
    if namespace == '':
        raise Exception('Must have a namespace')

    if not re.match(r'^ipa\.[0-9A-Za-z_]+', namespace):
        raise Exception('Namespace must be of the form "ipa.{pipeline_name}"')

def ValidateInterfaces(interfaces):
    # Validate presence of main interface
    intf = GetMainInterface(interfaces)
    if intf is None:
        raise Exception('Must have main IPA interface')

    # Validate presence of event interface
    event = GetEventInterface(interfaces)
    if intf is None:
        raise Exception('Must have event IPA interface')

    # Validate required main interface functions
    f_init  = [x for x in intf.methods if x.mojom_name == 'init']
    f_start = [x for x in intf.methods if x.mojom_name == 'start']
    f_stop  = [x for x in intf.methods if x.mojom_name == 'stop']

    ValidateSingleLength(f_init, 'init()', False)
    ValidateSingleLength(f_start, 'start()', False)
    ValidateSingleLength(f_stop, 'stop()', False)

    f_stop  = f_stop[0]

    # No need to validate init() and start() as they are customizable

    # Validate parameters to stop()
    ValidateZeroLength(f_stop.parameters, 'input parameter to stop()')
    ValidateZeroLength(f_stop.parameters, 'output parameter from stop()')

    # Validate that event interface has at least one event
    if len(event.methods) < 1:
        raise Exception('Event interface must have at least one event')

    # Validate that all async methods don't have return values
    intf_methods_async = [x for x in intf.methods if IsAsync(x)]
    for method in intf_methods_async:
        ValidateZeroLength(method.response_parameters,
                           f'{method.mojom_name} response parameters', False)

    event_methods_async = [x for x in event.methods if IsAsync(x)]
    for method in event_methods_async:
        ValidateZeroLength(method.response_parameters,
                           f'{method.mojom_name} response parameters', False)

class Generator(generator.Generator):
    @staticmethod
    def GetTemplatePrefix():
        return 'libcamera_templates'

    def GetFilters(self):
        libcamera_filters = {
            'all_types': GetAllTypes,
            'bit_width': BitWidth,
            'byte_width' : ByteWidthFromCppType,
            'cap': Capitalize,
            'choose': Choose,
            'comma_sep': CommaSep,
            'default_value': GetDefaultValue,
            'has_default_fields': HasDefaultFields,
            'has_fd': HasFd,
            'is_async': IsAsync,
            'is_array': IsArray,
            'is_controls': IsControls,
            'is_enum': IsEnum,
            'is_enum_scoped': IsEnumScoped,
            'is_fd': IsFd,
            'is_flags': IsFlags,
            'is_map': IsMap,
            'is_plain_struct': IsPlainStruct,
            'is_pod': IsPod,
            'is_scoped': IsScoped,
            'is_str': IsStr,
            'method_input_has_fd': MethodInputHasFd,
            'method_output_has_fd': MethodOutputHasFd,
            'method_param_names': MethodParamNames,
            'method_param_inputs': MethodParamInputs,
            'method_param_outputs': MethodParamOutputs,
            'method_parameters': MethodParameters,
            'method_return_value': MethodReturnValue,
            'name': GetNameForElement,
            'name_full': GetFullNameForElement,
            'needs_control_serializer': NeedsControlSerializer,
            'params_comma_sep': ParamsCommaSep,
            'with_default_values': WithDefaultValues,
            'with_fds': WithFds,
        }
        return libcamera_filters

    def _GetJinjaExports(self):
        return {
            'cmd_enum_name': '_%sCmd' % self.module_name,
            'cmd_event_enum_name': '_%sEventCmd' % self.module_name,
            'consts': self.module.constants,
            'enums': self.module.enums,
            'has_array': len([x for x in self.module.kinds.keys() if x[0] == 'a']) > 0,
            'has_map': len([x for x in self.module.kinds.keys() if x[0] == 'm']) > 0,
            'has_string': len([x for x in self.module.kinds.keys() if x[0] == 's']) > 0,
            'has_namespace': self.module.mojom_namespace != '',
            'interface_event': GetEventInterface(self.module.interfaces),
            'interface_main': GetMainInterface(self.module.interfaces),
            'interface_name': 'IPA%sInterface' % self.module_name,
            'module_name': ModuleName(self.module.path),
            'namespace': self.module.mojom_namespace.split('.'),
            'namespace_str': self.module.mojom_namespace.replace('.', '::') if
                             self.module.mojom_namespace is not None else '',
            'proxy_name': 'IPAProxy%s' % self.module_name,
            'proxy_worker_name': 'IPAProxy%sWorker' % self.module_name,
            'structs_nonempty': [x for x in self.module.structs if len(x.fields) > 0],
        }

    def _GetJinjaExportsForCore(self):
        return {
            'consts': self.module.constants,
            'enums_gen_header': [x for x in self.module.enums if x.attributes is None or 'skipHeader' not in x.attributes],
            'has_array': len([x for x in self.module.kinds.keys() if x[0] == 'a']) > 0,
            'has_map': len([x for x in self.module.kinds.keys() if x[0] == 'm']) > 0,
            'has_string': len([x for x in self.module.kinds.keys() if x[0] == 's']) > 0,
            'structs_gen_header': [x for x in self.module.structs if x.attributes is None or 'skipHeader' not in x.attributes],
            'structs_gen_serializer': [x for x in self.module.structs if x.attributes is None or 'skipSerdes' not in x.attributes],
        }

    @UseJinja('core_ipa_interface.h.tmpl')
    def _GenerateCoreHeader(self):
        return self._GetJinjaExportsForCore()

    @UseJinja('core_ipa_serializer.h.tmpl')
    def _GenerateCoreSerializer(self):
        return self._GetJinjaExportsForCore()

    @UseJinja('module_ipa_interface.h.tmpl')
    def _GenerateDataHeader(self):
        return self._GetJinjaExports()

    @UseJinja('module_ipa_serializer.h.tmpl')
    def _GenerateSerializer(self):
        return self._GetJinjaExports()

    @UseJinja('module_ipa_proxy.cpp.tmpl')
    def _GenerateProxyCpp(self):
        return self._GetJinjaExports()

    @UseJinja('module_ipa_proxy.h.tmpl')
    def _GenerateProxyHeader(self):
        return self._GetJinjaExports()

    @UseJinja('module_ipa_proxy_worker.cpp.tmpl')
    def _GenerateProxyWorker(self):
        return self._GetJinjaExports()

    def GenerateFiles(self, unparsed_args):
        parser = argparse.ArgumentParser()
        parser.add_argument('--libcamera_generate_core_header',     action='store_true')
        parser.add_argument('--libcamera_generate_core_serializer', action='store_true')
        parser.add_argument('--libcamera_generate_header',          action='store_true')
        parser.add_argument('--libcamera_generate_serializer',      action='store_true')
        parser.add_argument('--libcamera_generate_proxy_cpp',       action='store_true')
        parser.add_argument('--libcamera_generate_proxy_h',         action='store_true')
        parser.add_argument('--libcamera_generate_proxy_worker',    action='store_true')
        parser.add_argument('--libcamera_output_path')
        args = parser.parse_args(unparsed_args)

        if not args.libcamera_generate_core_header and \
           not args.libcamera_generate_core_serializer:
            ValidateNamespace(self.module.mojom_namespace)
            ValidateInterfaces(self.module.interfaces)
            self.module_name = ModuleClassName(self.module)

        fileutil.EnsureDirectoryExists(os.path.dirname(args.libcamera_output_path))

        gen_funcs = [
                [args.libcamera_generate_core_header,     self._GenerateCoreHeader],
                [args.libcamera_generate_core_serializer, self._GenerateCoreSerializer],
                [args.libcamera_generate_header,          self._GenerateDataHeader],
                [args.libcamera_generate_serializer,      self._GenerateSerializer],
                [args.libcamera_generate_proxy_cpp,       self._GenerateProxyCpp],
                [args.libcamera_generate_proxy_h,         self._GenerateProxyHeader],
                [args.libcamera_generate_proxy_worker,    self._GenerateProxyWorker],
        ]

        for pair in gen_funcs:
            if pair[0]:
                self.Write(pair[1](), args.libcamera_output_path)



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2022, Paul Elder <paul.elder@ideasonboard.com>
#
# Base class for a parser for a specific format of config file

class Parser(object):
    def __init__(self):
        pass

    # @brief Parse a config file into a config dict
    # @details The config dict shall have one key 'general' with a dict value
    #          for general configuration options, and all other entries shall
    #          have the module as the key with its configuration options (as a
    #          dict) as the value. The config dict shall prune entries that are
    #          for modules that are not in @a modules.
    # @param config (str) Path to config file
    # @param modules (list) List of modules
    # @return (dict, list) Configuration and list of modules to disable
    def parse(self, config_file: str, modules: list) -> (dict, list):
        raise NotImplementedError



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2022, Paul Elder <paul.elder@ideasonboard.com>

from libtuning.modules.lsc.lsc import LSC
from libtuning.modules.lsc.raspberrypi import ALSCRaspberryPi
from libtuning.modules.lsc.rkisp1 import LSCRkISP1



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
#
# camera tuning tool for GEQ (green equalisation)

from ctt_tools import *
import matplotlib.pyplot as plt
import scipy.optimize as optimize


"""
Uses green differences in macbeth patches to fit green equalisation threshold
model. Ideally, all macbeth chart centres would fall below the threshold as
these should be corrected by geq.
"""
def geq_fit(Cam, plot):
    imgs = Cam.imgs
    """
    green equalisation to mitigate mazing.
    Fits geq model by looking at difference
    between greens in macbeth patches
    """
    geqs = np.array([geq(Cam, Img)*Img.againQ8_norm for Img in imgs])
    Cam.log += '\nProcessed all images'
    geqs = geqs.reshape((-1, 2))
    """
    data is sorted by green difference and top half is selected since higher
    green difference data define the decision boundary.
    """
    geqs = np.array(sorted(geqs, key=lambda r: np.abs((r[1]-r[0])/r[0])))

    length = len(geqs)
    g0 = geqs[length//2:, 0]
    g1 = geqs[length//2:, 1]
    gdiff = np.abs(g0-g1)
    """
    find linear fit by minimising asymmetric least square errors
    in order to cover most of the macbeth images.
    the philosophy here is that every macbeth patch should fall within the
    threshold, hence the upper bound approach
    """
    def f(params):
        m, c = params
        a = gdiff - (m*g0+c)
        """
        asymmetric square error returns:
            1.95 * a**2 if a is positive
            0.05 * a**2 if a is negative
        """
        return(np.sum(a**2+0.95*np.abs(a)*a))

    initial_guess = [0.01, 500]
    """
    Nelder-Mead is usually not the most desirable optimisation method
    but has been chosen here due to its robustness to undifferentiability
    (is that a word?)
    """
    result = optimize.minimize(f, initial_guess, method='Nelder-Mead')
    """
    need to check if the fit worked correectly
    """
    if result.success:
        slope, offset = result.x
        Cam.log += '\nFit result: slope = {:.5f} '.format(slope)
        Cam.log += 'offset = {}'.format(int(offset))
        """
        optional plotting code
        """
        if plot:
            x = np.linspace(max(g0)*1.1, 100)
            y = slope*x + offset
            plt.title('GEQ Asymmetric \'Upper Bound\' Fit')
            plt.plot(x, y, color='red', ls='--', label='fit')
            plt.scatter(g0, gdiff, color='b', label='data')
            plt.ylabel('Difference in green channels')
            plt.xlabel('Green value')

        """
        This upper bound asymmetric gives correct order of magnitude values.
        The pipeline approximates a 1st derivative of a gaussian with some
        linear piecewise functions, introducing arbitrary cutoffs. For
        pessimistic geq, the model parameters have been increased by a
        scaling factor/constant.

        Feel free to tune these or edit the json files directly if you
        belive there are still mazing effects left (threshold too low) or if you
        think it is being overcorrected (threshold too high).
        We have gone for a one size fits most approach that will produce
        acceptable results in most applications.
        """
        slope *= 1.5
        offset += 201
        Cam.log += '\nFit after correction factors: slope = {:.5f}'.format(slope)
        Cam.log += ' offset = {}'.format(int(offset))
        """
        clamp offset at 0 due to pipeline considerations
        """
        if offset < 0:
            Cam.log += '\nOffset raised to 0'
            offset = 0
        """
        optional plotting code
        """
        if plot:
            y2 = slope*x + offset
            plt.plot(x, y2, color='green', ls='--', label='scaled fit')
            plt.grid()
            plt.legend()
            plt.show()

        """
    the case where for some reason the fit didn't work correctly

    Transpose data and then least squares linear fit. Transposing data
    makes it robust to many patches where green difference is the same
    since they only contribute to one error minimisation, instead of dragging
    the entire linear fit down.
    """

    else:
        print('\nError! Couldn\'t fit asymmetric lest squares')
        print(result.message)
        Cam.log += '\nWARNING: Asymmetric least squares fit failed! '
        Cam.log += 'Standard fit used could possibly lead to worse results'
        fit = np.polyfit(gdiff, g0, 1)
        offset, slope = -fit[1]/fit[0], 1/fit[0]
        Cam.log += '\nFit result: slope = {:.5f} '.format(slope)
        Cam.log += 'offset = {}'.format(int(offset))
        """
        optional plotting code
        """
        if plot:
            x = np.linspace(max(g0)*1.1, 100)
            y = slope*x + offset
            plt.title('GEQ Linear Fit')
            plt.plot(x, y, color='red', ls='--', label='fit')
            plt.scatter(g0, gdiff, color='b', label='data')
            plt.ylabel('Difference in green channels')
            plt.xlabel('Green value')
        """
        Scaling factors (see previous justification)
        The model here will not be an upper bound so scaling factors have
        been increased.
        This method of deriving geq model parameters is extremely arbitrary
        and undesirable.
        """
        slope *= 2.5
        offset += 301
        Cam.log += '\nFit after correction factors: slope = {:.5f}'.format(slope)
        Cam.log += ' offset = {}'.format(int(offset))

        if offset < 0:
            Cam.log += '\nOffset raised to 0'
            offset = 0

        """
        optional plotting code
        """
        if plot:
            y2 = slope*x + offset
            plt.plot(x, y2, color='green', ls='--', label='scaled fit')
            plt.legend()
            plt.grid()
            plt.show()

    return round(slope, 5), int(offset)


""""
Return green channels of macbeth patches
returns g0, g1 where
> g0 is green next to red
> g1 is green next to blue
"""
def geq(Cam, Img):
    Cam.log += '\nProcessing image {}'.format(Img.name)
    patches = [Img.patches[i] for i in Img.order][1:3]
    g_patches = np.array([(np.mean(patches[0][i]), np.mean(patches[1][i])) for i in range(24)])
    Cam.log += '\n'
    return(g_patches)



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
#
# camera tuning tool for ALSC (auto lens shading correction)

from ctt_image_load import *
import matplotlib.pyplot as plt
from matplotlib import cm
from mpl_toolkits.mplot3d import Axes3D


"""
preform alsc calibration on a set of images
"""
def alsc_all(Cam, do_alsc_colour, plot, grid_size=(16, 12), max_gain=8.0):
    imgs_alsc = Cam.imgs_alsc
    grid_w, grid_h = grid_size
    """
    create list of colour temperatures and associated calibration tables
    """
    list_col = []
    list_cr = []
    list_cb = []
    list_cg = []
    for Img in imgs_alsc:
        col, cr, cb, cg, size = alsc(Cam, Img, do_alsc_colour, plot, grid_size=grid_size, max_gain=max_gain)
        list_col.append(col)
        list_cr.append(cr)
        list_cb.append(cb)
        list_cg.append(cg)
        Cam.log += '\n'
    Cam.log += '\nFinished processing images'
    w, h, dx, dy = size
    Cam.log += '\nChannel dimensions: w = {}  h = {}'.format(int(w), int(h))
    Cam.log += '\n16x12 grid rectangle size: w = {} h = {}'.format(dx, dy)

    """
    convert to numpy array for data manipulation
    """
    list_col = np.array(list_col)
    list_cr = np.array(list_cr)
    list_cb = np.array(list_cb)
    list_cg = np.array(list_cg)

    cal_cr_list = []
    cal_cb_list = []

    """
    only do colour calculations if required
    """
    if do_alsc_colour:
        Cam.log += '\nALSC colour tables'
        for ct in sorted(set(list_col)):
            Cam.log += '\nColour temperature: {} K'.format(ct)
            """
            average tables for the same colour temperature
            """
            indices = np.where(list_col == ct)
            ct = int(ct)
            t_r = np.mean(list_cr[indices], axis=0)
            t_b = np.mean(list_cb[indices], axis=0)
            """
            force numbers to be stored to 3dp.... :(
            """
            t_r = np.where((100*t_r) % 1 <= 0.05, t_r+0.001, t_r)
            t_b = np.where((100*t_b) % 1 <= 0.05, t_b+0.001, t_b)
            t_r = np.where((100*t_r) % 1 >= 0.95, t_r-0.001, t_r)
            t_b = np.where((100*t_b) % 1 >= 0.95, t_b-0.001, t_b)
            t_r = np.round(t_r, 3)
            t_b = np.round(t_b, 3)
            r_corners = (t_r[0], t_r[grid_w - 1], t_r[-1], t_r[-grid_w])
            b_corners = (t_b[0], t_b[grid_w - 1], t_b[-1], t_b[-grid_w])
            middle_pos = (grid_h // 2 - 1) * grid_w + grid_w - 1
            r_cen = t_r[middle_pos]+t_r[middle_pos + 1]+t_r[middle_pos + grid_w]+t_r[middle_pos + grid_w + 1]
            r_cen = round(r_cen/4, 3)
            b_cen = t_b[middle_pos]+t_b[middle_pos + 1]+t_b[middle_pos + grid_w]+t_b[middle_pos + grid_w + 1]
            b_cen = round(b_cen/4, 3)
            Cam.log += '\nRed table corners: {}'.format(r_corners)
            Cam.log += '\nRed table centre: {}'.format(r_cen)
            Cam.log += '\nBlue table corners: {}'.format(b_corners)
            Cam.log += '\nBlue table centre: {}'.format(b_cen)
            cr_dict = {
                'ct': ct,
                'table': list(t_r)
            }
            cb_dict = {
                'ct': ct,
                'table': list(t_b)
            }
            cal_cr_list.append(cr_dict)
            cal_cb_list.append(cb_dict)
            Cam.log += '\n'
    else:
        cal_cr_list, cal_cb_list = None, None

    """
    average all values for luminance shading and return one table for all temperatures
    """
    lum_lut = np.mean(list_cg, axis=0)
    lum_lut = np.where((100*lum_lut) % 1 <= 0.05, lum_lut+0.001, lum_lut)
    lum_lut = np.where((100*lum_lut) % 1 >= 0.95, lum_lut-0.001, lum_lut)
    lum_lut = list(np.round(lum_lut, 3))

    """
    calculate average corner for lsc gain calculation further on
    """
    corners = (lum_lut[0], lum_lut[15], lum_lut[-1], lum_lut[-16])
    Cam.log += '\nLuminance table corners: {}'.format(corners)
    l_cen = lum_lut[5*16+7]+lum_lut[5*16+8]+lum_lut[6*16+7]+lum_lut[6*16+8]
    l_cen = round(l_cen/4, 3)
    Cam.log += '\nLuminance table centre: {}'.format(l_cen)
    av_corn = np.sum(corners)/4

    return cal_cr_list, cal_cb_list, lum_lut, av_corn


"""
calculate g/r and g/b for 32x32 points arranged in a grid for a single image
"""
def alsc(Cam, Img, do_alsc_colour, plot=False, grid_size=(16, 12), max_gain=8.0):
    Cam.log += '\nProcessing image: ' + Img.name
    grid_w, grid_h = grid_size
    """
    get channel in correct order
    """
    channels = [Img.channels[i] for i in Img.order]
    """
    calculate size of single rectangle.
    -(-(w-1)//32) is a ceiling division. w-1 is to deal robustly with the case
    where w is a multiple of 32.
    """
    w, h = Img.w/2, Img.h/2
    dx, dy = int(-(-(w-1)//grid_w)), int(-(-(h-1)//grid_h))
    """
    average the green channels into one
    """
    av_ch_g = np.mean((channels[1:3]), axis=0)
    if do_alsc_colour:
        """
        obtain grid_w x grid_h grid of intensities for each channel and subtract black level
        """
        g = get_grid(av_ch_g, dx, dy, grid_size) - Img.blacklevel_16
        r = get_grid(channels[0], dx, dy, grid_size) - Img.blacklevel_16
        b = get_grid(channels[3], dx, dy, grid_size) - Img.blacklevel_16
        """
        calculate ratios as 32 bit in order to be supported by medianBlur function
        """
        cr = np.reshape(g/r, (grid_h, grid_w)).astype('float32')
        cb = np.reshape(g/b, (grid_h, grid_w)).astype('float32')
        cg = np.reshape(1/g, (grid_h, grid_w)).astype('float32')
        """
        median blur to remove peaks and save as float 64
        """
        cr = cv2.medianBlur(cr, 3).astype('float64')
        cr = cr/np.min(cr)  # gain tables are easier for humans to read if the minimum is 1.0
        cb = cv2.medianBlur(cb, 3).astype('float64')
        cb = cb/np.min(cb)
        cg = cv2.medianBlur(cg, 3).astype('float64')
        cg = cg/np.min(cg)
        cg = [min(v, max_gain) for v in cg.flatten()]  # never exceed the max luminance gain

        """
        debugging code showing 2D surface plot of vignetting. Quite useful for
        for sanity check
        """
        if plot:
            hf = plt.figure(figsize=(8, 8))
            ha = hf.add_subplot(311, projection='3d')
            """
            note Y is plotted as -Y so plot has same axes as image
            """
            X, Y = np.meshgrid(range(grid_w), range(grid_h))
            ha.plot_surface(X, -Y, cr, cmap=cm.coolwarm, linewidth=0)
            ha.set_title('ALSC Plot\nImg: {}\n\ncr'.format(Img.str))
            hb = hf.add_subplot(312, projection='3d')
            hb.plot_surface(X, -Y, cb, cmap=cm.coolwarm, linewidth=0)
            hb.set_title('cb')
            hc = hf.add_subplot(313, projection='3d')
            hc.plot_surface(X, -Y, cg, cmap=cm.coolwarm, linewidth=0)
            hc.set_title('g')
            # print(Img.str)
            plt.show()

        return Img.col, cr.flatten(), cb.flatten(), cg, (w, h, dx, dy)

    else:
        """
        only perform calculations for luminance shading
        """
        g = get_grid(av_ch_g, dx, dy, grid_size) - Img.blacklevel_16
        cg = np.reshape(1/g, (grid_h, grid_w)).astype('float32')
        cg = cv2.medianBlur(cg, 3).astype('float64')
        cg = cg/np.min(cg)
        cg = [min(v, max_gain) for v in cg.flatten()]  # never exceed the max luminance gain

        if plot:
            hf = plt.figure(figssize=(8, 8))
            ha = hf.add_subplot(1, 1, 1, projection='3d')
            X, Y = np.meashgrid(range(grid_w), range(grid_h))
            ha.plot_surface(X, -Y, cg, cmap=cm.coolwarm, linewidth=0)
            ha.set_title('ALSC Plot (Luminance only!)\nImg: {}\n\ncg').format(Img.str)
            plt.show()

        return Img.col, None, None, cg.flatten(), (w, h, dx, dy)


"""
Compresses channel down to a grid of the requested size
"""
def get_grid(chan, dx, dy, grid_size):
    grid_w, grid_h = grid_size
    grid = []
    """
    since left and bottom border will not necessarily have rectangles of
    dimension dx x dy, the 32nd iteration has to be handled separately.
    """
    for i in range(grid_h - 1):
        for j in range(grid_w - 1):
            grid.append(np.mean(chan[dy*i:dy*(1+i), dx*j:dx*(1+j)]))
        grid.append(np.mean(chan[dy*i:dy*(1+i), (grid_w - 1)*dx:]))
    for j in range(grid_w - 1):
        grid.append(np.mean(chan[(grid_h - 1)*dy:, dx*j:dx*(1+j)]))
    grid.append(np.mean(chan[(grid_h - 1)*dy:, (grid_w - 1)*dx:]))
    """
    return as np.array, ready for further manipulation
    """
    return np.array(grid)


"""
obtains sigmas for red and blue, effectively a measure of the 'error'
"""
def get_sigma(Cam, cal_cr_list, cal_cb_list, grid_size):
    Cam.log += '\nCalculating sigmas'
    """
    provided colour alsc tables were generated for two different colour
    temperatures sigma is calculated by comparing two calibration temperatures
    adjacent in colour space
    """
    """
    create list of colour temperatures
    """
    cts = [cal['ct'] for cal in cal_cr_list]
    # print(cts)
    """
    calculate sigmas for each adjacent cts and return worst one
    """
    sigma_rs = []
    sigma_bs = []
    for i in range(len(cts)-1):
        sigma_rs.append(calc_sigma(cal_cr_list[i]['table'], cal_cr_list[i+1]['table'], grid_size))
        sigma_bs.append(calc_sigma(cal_cb_list[i]['table'], cal_cb_list[i+1]['table'], grid_size))
        Cam.log += '\nColour temperature interval {} - {} K'.format(cts[i], cts[i+1])
        Cam.log += '\nSigma red: {}'.format(sigma_rs[-1])
        Cam.log += '\nSigma blue: {}'.format(sigma_bs[-1])

    """
    return maximum sigmas, not necessarily from the same colour temperature
    interval
    """
    sigma_r = max(sigma_rs) if sigma_rs else 0.005
    sigma_b = max(sigma_bs) if sigma_bs else 0.005
    Cam.log += '\nMaximum sigmas: Red = {} Blue = {}'.format(sigma_r, sigma_b)

    # print(sigma_rs, sigma_bs)
    # print(sigma_r, sigma_b)
    return sigma_r, sigma_b


"""
calculate sigma from two adjacent gain tables
"""
def calc_sigma(g1, g2, grid_size):
    grid_w, grid_h = grid_size
    """
    reshape into 16x12 matrix
    """
    g1 = np.reshape(g1, (grid_h, grid_w))
    g2 = np.reshape(g2, (grid_h, grid_w))
    """
    apply gains to gain table
    """
    gg = g1/g2
    if np.mean(gg) < 1:
        gg = 1/gg
    """
    for each internal patch, compute average difference between it and its 4
    neighbours, then append to list
    """
    diffs = []
    for i in range(grid_h - 2):
        for j in range(grid_w - 2):
            """
            note indexing is incremented by 1 since all patches on borders are
            not counted
            """
            diff = np.abs(gg[i+1][j+1]-gg[i][j+1])
            diff += np.abs(gg[i+1][j+1]-gg[i+2][j+1])
            diff += np.abs(gg[i+1][j+1]-gg[i+1][j])
            diff += np.abs(gg[i+1][j+1]-gg[i+1][j+2])
            diffs.append(diff/4)

    """
    return mean difference
    """
    mean_diff = np.mean(diffs)
    return(np.round(mean_diff, 5))



#!/usr/bin/env python
# Copyright 2019 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
#
# This utility concatenates several files into one. On Unix-like systems
# it is equivalent to:
#   cat file1 file2 file3 ...files... > target
#
# The reason for writing a separate utility is that 'cat' is not available
# on all supported build platforms, but Python is, and hence this provides
# us with an easy and uniform way of doing this on all platforms.

# for py2/py3 compatibility
from __future__ import print_function

import optparse
import sys


def Concatenate(filenames):
  """Concatenate files.

  Args:
    files: Array of file names.
           The last name is the target; all earlier ones are sources.

  Returns:
    True, if the operation was successful.
  """
  if len(filenames) < 2:
    print("An error occurred generating %s:\nNothing to do." % filenames[-1])
    return False

  try:
    with open(filenames[-1], "wb") as target:
      for filename in filenames[:-1]:
        with open(filename, "rb") as current:
          target.write(current.read())
    return True
  except IOError as e:
    print("An error occurred when writing %s:\n%s" % (filenames[-1], e))
    return False


def main():
  parser = optparse.OptionParser()
  parser.set_usage("""Concatenate several files into one.
      Equivalent to: cat file1 ... > target.""")
  (_options, args) = parser.parse_args()
  sys.exit(0 if Concatenate(args) else 1)


if __name__ == "__main__":
  main()



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2024, Paul Elder <paul.elder@ideasonboard.com>

from libtuning.modules.ccm.ccm import CCM
from libtuning.modules.ccm.rkisp1 import CCMRkISP1



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2022, Ideas on Board Oy
#
# Generate color space conversion table coefficients with configurable
# fixed-point precision

import argparse
import enum
import numpy as np
import sys


encodings = {
    'rec601': [
        [  0.299,          0.587,          0.114         ],
        [ -0.299 / 1.772, -0.587 / 1.772,  0.886 / 1.772 ],
        [  0.701 / 1.402, -0.587 / 1.402, -0.114 / 1.402 ]
    ],
    'rec709': [
        [  0.2126,           0.7152,           0.0722          ],
        [ -0.2126 / 1.8556, -0.7152 / 1.8556,  0.9278 / 1.8556 ],
        [  0.7874 / 1.5748, -0.7152 / 1.5748, -0.0722 / 1.5748 ]
    ],
    'rec2020': [
        [  0.2627,           0.6780,           0.0593          ],
        [ -0.2627 / 1.8814, -0.6780 / 1.8814,  0.9407 / 1.8814 ],
        [  0.7373 / 1.4746, -0.6780 / 1.4746, -0.0593 / 1.4746 ],
    ],
    'smpte240m': [
        [  0.2122,           0.7013,           0.0865          ],
        [ -0.2122 / 1.8270, -0.7013 / 1.8270,  0.9135 / 1.8270 ],
        [  0.7878 / 1.5756, -0.7013 / 1.5756, -0.0865 / 1.5756 ],
    ],
}


class Precision(object):
    def __init__(self, precision):
        if precision[0].upper() != 'Q':
            raise RuntimeError(f'Invalid precision `{precision}`')
        prec = precision[1:].split('.')
        if len(prec) != 2:
            raise RuntimeError(f'Invalid precision `{precision}`')

        self.__prec = [int(v) for v in prec]

    @property
    def integer(self):
        return self.__prec[0]

    @property
    def fractional(self):
        return self.__prec[1]

    @property
    def total(self):
        # Add 1 for the sign bit
        return self.__prec[0] + self.__prec[1] + 1


class Quantization(enum.Enum):
    FULL = 0
    LIMITED = 1


def scale_coeff(coeff, quantization, luma):
    """Scale a coefficient to the output range dictated by the quantization.

    Parameters
    ----------
    coeff : float
        The CSC matrix coefficient to scale
    quantization : Quantization
        The quantization, either FULL or LIMITED
    luma : bool
        True if the coefficient corresponds to a luma value, False otherwise
    """

    # Assume the input range is 8 bits. The output range is set by the
    # quantization and differs between luma and chrome components for limited
    # range.
    in_range = 255 - 0
    if quantization == Quantization.FULL:
        out_range = 255 - 0
    elif luma:
        out_range = 235 - 16
    else:
        out_range = 240 - 16

    return coeff * out_range / in_range


def round_array(values):
    """Round a list of signed floating point values to the closest integer while
    preserving the (rounded) value of the sum of all elements.
    """

    # Calculate the rounding error as the difference between the rounded sum of
    # values and the sum of rounded values. This is by definition an integer
    # (positive or negative), which indicates how many values will need to be
    # 'flipped' to the opposite rounding.
    rounded_values = [round(value) for value in values]
    sum_values = round(sum(values))
    sum_error = sum_values - sum(rounded_values)

    if sum_error == 0:
        return rounded_values

    # The next step is to distribute the error among the values, in a way that
    # will minimize the relative error introduced in individual values. We
    # extend the values list with the rounded value and original index for each
    # element, and sort by rounding error. Then we modify the elements with the
    # highest or lowest error, depending on whether the sum error is negative
    # or positive.

    values = [[value, round(value), index] for index, value in enumerate(values)]
    values.sort(key=lambda v: v[1] - v[0])

    # It could also be argued that the key for the sort order should not be the
    # absolute rouding error but the relative error, as the impact of identical
    # rounding errors will differ for coefficients with widely different values.
    # This is a topic for further research.
    #
    # values.sort(key=lambda v: (v[1] - v[0]) / abs(v[0]))

    if sum_error > 0:
        for i in range(sum_error):
            values[i][1] += 1
    else:
        for i in range(-sum_error):
            values[len(values) - i - 1][1] -= 1

    # Finally, sort back by index, make sure the total rounding error is now 0,
    # and return the rounded values.
    values.sort(key=lambda v: v[2])
    values = [value[1] for value in values]
    assert(sum(values) == sum_values)

    return values


def main(argv):

    # Parse command line arguments.
    parser = argparse.ArgumentParser(
        description='Generate color space conversion table coefficients with '
        'configurable fixed-point precision.'
    )
    parser.add_argument('--invert', '-i', action='store_true',
                        help='Invert the color space conversion (YUV -> RGB)')
    parser.add_argument('--precision', '-p', default='Q1.7',
                        help='The output fixed point precision in Q notation (sign bit excluded)')
    parser.add_argument('--quantization', '-q', choices=['full', 'limited'],
                        default='limited', help='Quantization range')
    parser.add_argument('encoding', choices=encodings.keys(), help='YCbCr encoding')
    args = parser.parse_args(argv[1:])

    try:
        precision = Precision(args.precision)
    except Exception:
        print(f'Invalid precision `{args.precision}`')
        return 1

    encoding = encodings[args.encoding]
    quantization = Quantization[args.quantization.upper()]

    # Scale and round the encoding coefficients based on the precision and
    # quantization range.
    luma = True
    scaled_coeffs = []
    for line in encoding:
        line = [scale_coeff(coeff, quantization, luma) for coeff in line]
        scaled_coeffs.append(line)
        luma = False

    if args.invert:
        scaled_coeffs = np.linalg.inv(scaled_coeffs)

    rounded_coeffs = []
    for line in scaled_coeffs:
        line = [coeff * (1 << precision.fractional) for coeff in line]
        # For the RGB to YUV conversion, use a rounding method that preserves
        # the rounded sum of each line to avoid biases and overflow, as the sum
        # of luma and chroma coefficients should be 1.0 and 0.0 respectively
        # (in full range). For the YUV to RGB conversion, there is no such
        # constraint, so use simple rounding.
        if args.invert:
            line = [round(coeff) for coeff in line]
        else:
            line = round_array(line)

        # Convert coefficients to the number of bits selected by the precision.
        # Negative values will be turned into positive integers using 2's
        # complement.
        line = [coeff & ((1 << precision.total) - 1) for coeff in line]
        rounded_coeffs.append(line)

    # Print the result as C code.
    nbits = 1 << (precision.total - 1).bit_length()
    nbytes = nbits // 4
    print(f'static const u{nbits} {"yuv2rgb" if args.invert else "rgb2yuv"}_{args.encoding}_{quantization.name.lower()}_coeffs[] = {{')

    for line in rounded_coeffs:
        line = [f'0x{coeff:0{nbytes}x}' for coeff in line]

        print(f'\t{", ".join(line)},')

    print('};')

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2024, Paul Elder <paul.elder@ideasonboard.com>
# Copyright (C) 2024, Ideas on Board
#
# Base Ccm tuning module

from ..module import Module

from libtuning.ctt_ccm import ccm
import logging

logger = logging.getLogger(__name__)


class CCM(Module):
    type = 'ccm'
    hr_name = 'CCM (Base)'
    out_name = 'GenericCCM'

    def __init__(self, debug: list):
        super().__init__()

        self.debug = debug

    def do_calibration(self, images):
        logger.info('Starting CCM calibration')

        imgs = [img for img in images if img.macbeth is not None]

        # todo: Take LSC calibration results into account.
        cal_cr_list = None
        cal_cb_list = None

        try:
            ccms = ccm(imgs, cal_cr_list, cal_cb_list)
        except ArithmeticError:
            logger.error('CCM calibration failed')
            return None

        return ccms



# SPDX-License-Identifier: BSD-2-Clause

import re
import sys
import os

if len(sys.argv) != 2:
    print("Usage: {} <infile>".format(sys.argv[0]))
    sys.exit()

infile = sys.argv[1]
insplit = os.path.splitext(infile)
outfile = insplit[0] + '_parsed' + insplit[1]

frame_re = re.compile(r'frame (\d+) started')

delays = {
    'Analogue Gain': 1,
    'Exposure': 2,
    'Vertical Blanking': 2
}

ctrl_action = {
    'Write': {},
    'Get': {},
    'Queue': {},
    'No-op': {}
}

ctrl_re = {
    'Write': re.compile(r'Setting (.*?) to (\d+) at index (\d+)'),
    'No-op': re.compile(r'Queue is empty, (.*?) (.*?) (.*?)'),
    'Get': re.compile(r'Reading (.*?) to (\d+) at index (\d+)'),
    'Queue': re.compile(r'Queuing (.*?) to (\d+) at index (\d+)')
}

frame_num = -1

max_delay = 0
for k, d in delays.items():
    if max_delay < d:
        max_delay = d

with open(infile) as f:
    lines = f.readlines()

for line in lines:
    r = frame_re.search(line)
    if r:
        frame_num = int(r.group(1))

    for (key, re) in ctrl_re.items():
        r = re.search(line)
        if r:
            ctrl_action[key][(frame_num, r.group(1))] = (r.group(2), r.group(3))

with open(outfile, 'wt') as f:
    queueIndex = 1
    f.write('{:<10}{:<15}{:<12}{:<18}{}\n'.format('Frame', 'Action', 'Gain', 'Exposure', 'Vblank'))
    for frame in range(0, frame_num + 1):
        for (k, a) in ctrl_action.items():
            str = '{:<10}{:<10}'.format(frame, k)

            for c in delays.keys():
                # Tabulate all results
                str += '{:>5} {:<10}'.format(a[(frame, c)][0] if (frame, c) in a.keys() else '---',
                                             '[' + (a[(frame, c)][1] if (frame, c) in a.keys() else '-') + ']')

            f.write(str.strip() + '\n')

# Test the write -> get matches the set delay.
for (frame, c) in ctrl_action['Write'].keys():
    set_value = ctrl_action['Write'][(frame, c)][0]
    delay_frame = frame + delays[c]
    if (delay_frame <= frame_num):
        if (delay_frame, c) in ctrl_action['Get']:
            get_value = ctrl_action['Get'][(delay_frame, c)][0]
            if get_value != set_value:
                print('Error: {} written at frame {} to value {} != {} at frame {}'
                      .format(c, frame, set_value, get_value, delay_frame))
        else:
            print('Warning: {} written at frame {} to value {} did not get logged on frame {} - dropped frame?'
                  .format(c, frame, set_value, delay_frame))

# Test the queue -> write matches the set delay.
for (frame, c) in ctrl_action['Queue'].keys():
    set_value = ctrl_action['Queue'][(frame, c)][0]
    delay_frame = frame + max_delay - delays[c] + 1
    if (delay_frame <= frame_num):
        if (delay_frame, c) in ctrl_action['Write']:
            write_value = ctrl_action['Write'][(delay_frame, c)][0]
            if write_value != set_value:
                print('Info: {} queued at frame {} to value {} != {} written at frame {}'
                      ' - lagging behind or double queue on a single frame!'
                      .format(c, frame, set_value, write_value, delay_frame))
        else:
            print('Warning: {} queued at frame {} to value {} did not get logged on frame {} - dropped frame?'
                  .format(c, frame, set_value, delay_frame))

# Test the get -> write matches the set delay going backwards.
for (frame, c) in ctrl_action['Get'].keys():
    get_value = ctrl_action['Get'][(frame, c)][0]
    delay_frame = frame - delays[c]
    if (delay_frame >= 6):
        if (delay_frame, c) in ctrl_action['Write']:
            write_value = ctrl_action['Write'][(delay_frame, c)][0]
            if get_value != write_value:
                print('Info: {} got at frame {} to value {} != {} written at frame {}'
                      ' - lagging behind or double queue on a single frame!'
                      .format(c, frame, get_value, write_value, delay_frame))
        else:
            print('Warning: {} got at frame {} to value {} did not get written on frame {}'
                  .format(c, frame, get_value, delay_frame))



# Copyright 2014 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import importlib.util
import os.path
import sys
import unittest

def _GetDirAbove(dirname):
  """Returns the directory "above" this file containing |dirname| (which must
  also be "above" this file)."""
  path = os.path.abspath(__file__)
  while True:
    path, tail = os.path.split(path)
    assert tail
    if tail == dirname:
      return path

sys.path.insert(1, os.path.join(_GetDirAbove("mojo"), "third_party"))
from ply import lex

try:
  importlib.util.find_spec("mojom")
except ImportError:
  sys.path.append(os.path.join(_GetDirAbove("pylib"), "pylib"))
import mojom.parse.lexer

# This (monkey-patching LexToken to make comparison value-based) is evil, but
# we'll do it anyway. (I'm pretty sure ply's lexer never cares about comparing
# for object identity.)
def _LexTokenEq(self, other):
  return self.type == other.type and self.value == other.value and \
         self.lineno == other.lineno and self.lexpos == other.lexpos


setattr(lex.LexToken, '__eq__', _LexTokenEq)


def _MakeLexToken(token_type, value, lineno=1, lexpos=0):
  """Makes a LexToken with the given parameters. (Note that lineno is 1-based,
  but lexpos is 0-based.)"""
  rv = lex.LexToken()
  rv.type, rv.value, rv.lineno, rv.lexpos = token_type, value, lineno, lexpos
  return rv


def _MakeLexTokenForKeyword(keyword, **kwargs):
  """Makes a LexToken for the given keyword."""
  return _MakeLexToken(keyword.upper(), keyword.lower(), **kwargs)


class LexerTest(unittest.TestCase):
  """Tests |mojom.parse.lexer.Lexer|."""

  def __init__(self, *args, **kwargs):
    unittest.TestCase.__init__(self, *args, **kwargs)
    # Clone all lexer instances from this one, since making a lexer is slow.
    self._zygote_lexer = lex.lex(mojom.parse.lexer.Lexer("my_file.mojom"))

  def testValidKeywords(self):
    """Tests valid keywords."""
    self.assertEquals(
        self._SingleTokenForInput("handle"), _MakeLexTokenForKeyword("handle"))
    self.assertEquals(
        self._SingleTokenForInput("import"), _MakeLexTokenForKeyword("import"))
    self.assertEquals(
        self._SingleTokenForInput("module"), _MakeLexTokenForKeyword("module"))
    self.assertEquals(
        self._SingleTokenForInput("struct"), _MakeLexTokenForKeyword("struct"))
    self.assertEquals(
        self._SingleTokenForInput("union"), _MakeLexTokenForKeyword("union"))
    self.assertEquals(
        self._SingleTokenForInput("interface"),
        _MakeLexTokenForKeyword("interface"))
    self.assertEquals(
        self._SingleTokenForInput("enum"), _MakeLexTokenForKeyword("enum"))
    self.assertEquals(
        self._SingleTokenForInput("const"), _MakeLexTokenForKeyword("const"))
    self.assertEquals(
        self._SingleTokenForInput("true"), _MakeLexTokenForKeyword("true"))
    self.assertEquals(
        self._SingleTokenForInput("false"), _MakeLexTokenForKeyword("false"))
    self.assertEquals(
        self._SingleTokenForInput("default"),
        _MakeLexTokenForKeyword("default"))
    self.assertEquals(
        self._SingleTokenForInput("array"), _MakeLexTokenForKeyword("array"))
    self.assertEquals(
        self._SingleTokenForInput("map"), _MakeLexTokenForKeyword("map"))
    self.assertEquals(
        self._SingleTokenForInput("associated"),
        _MakeLexTokenForKeyword("associated"))

  def testValidIdentifiers(self):
    """Tests identifiers."""
    self.assertEquals(
        self._SingleTokenForInput("abcd"), _MakeLexToken("NAME", "abcd"))
    self.assertEquals(
        self._SingleTokenForInput("AbC_d012_"),
        _MakeLexToken("NAME", "AbC_d012_"))
    self.assertEquals(
        self._SingleTokenForInput("_0123"), _MakeLexToken("NAME", "_0123"))

  def testInvalidIdentifiers(self):
    with self.assertRaisesRegexp(
        mojom.parse.lexer.LexError,
        r"^my_file\.mojom:1: Error: Illegal character '\$'$"):
      self._TokensForInput("$abc")
    with self.assertRaisesRegexp(
        mojom.parse.lexer.LexError,
        r"^my_file\.mojom:1: Error: Illegal character '\$'$"):
      self._TokensForInput("a$bc")

  def testDecimalIntegerConstants(self):
    self.assertEquals(
        self._SingleTokenForInput("0"), _MakeLexToken("INT_CONST_DEC", "0"))
    self.assertEquals(
        self._SingleTokenForInput("1"), _MakeLexToken("INT_CONST_DEC", "1"))
    self.assertEquals(
        self._SingleTokenForInput("123"), _MakeLexToken("INT_CONST_DEC", "123"))
    self.assertEquals(
        self._SingleTokenForInput("10"), _MakeLexToken("INT_CONST_DEC", "10"))

  def testValidTokens(self):
    """Tests valid tokens (which aren't tested elsewhere)."""
    # Keywords tested in |testValidKeywords|.
    # NAME tested in |testValidIdentifiers|.
    self.assertEquals(
        self._SingleTokenForInput("@123"), _MakeLexToken("ORDINAL", "@123"))
    self.assertEquals(
        self._SingleTokenForInput("456"), _MakeLexToken("INT_CONST_DEC", "456"))
    self.assertEquals(
        self._SingleTokenForInput("0x01aB2eF3"),
        _MakeLexToken("INT_CONST_HEX", "0x01aB2eF3"))
    self.assertEquals(
        self._SingleTokenForInput("123.456"),
        _MakeLexToken("FLOAT_CONST", "123.456"))
    self.assertEquals(
        self._SingleTokenForInput("\"hello\""),
        _MakeLexToken("STRING_LITERAL", "\"hello\""))
    self.assertEquals(
        self._SingleTokenForInput("+"), _MakeLexToken("PLUS", "+"))
    self.assertEquals(
        self._SingleTokenForInput("-"), _MakeLexToken("MINUS", "-"))
    self.assertEquals(
        self._SingleTokenForInput("?"), _MakeLexToken("QSTN", "?"))
    self.assertEquals(
        self._SingleTokenForInput("="), _MakeLexToken("EQUALS", "="))
    self.assertEquals(
        self._SingleTokenForInput("=>"), _MakeLexToken("RESPONSE", "=>"))
    self.assertEquals(
        self._SingleTokenForInput("("), _MakeLexToken("LPAREN", "("))
    self.assertEquals(
        self._SingleTokenForInput(")"), _MakeLexToken("RPAREN", ")"))
    self.assertEquals(
        self._SingleTokenForInput("["), _MakeLexToken("LBRACKET", "["))
    self.assertEquals(
        self._SingleTokenForInput("]"), _MakeLexToken("RBRACKET", "]"))
    self.assertEquals(
        self._SingleTokenForInput("{"), _MakeLexToken("LBRACE", "{"))
    self.assertEquals(
        self._SingleTokenForInput("}"), _MakeLexToken("RBRACE", "}"))
    self.assertEquals(
        self._SingleTokenForInput("<"), _MakeLexToken("LANGLE", "<"))
    self.assertEquals(
        self._SingleTokenForInput(">"), _MakeLexToken("RANGLE", ">"))
    self.assertEquals(
        self._SingleTokenForInput(";"), _MakeLexToken("SEMI", ";"))
    self.assertEquals(
        self._SingleTokenForInput(","), _MakeLexToken("COMMA", ","))
    self.assertEquals(self._SingleTokenForInput("."), _MakeLexToken("DOT", "."))

  def _TokensForInput(self, input_string):
    """Gets a list of tokens for the given input string."""
    lexer = self._zygote_lexer.clone()
    lexer.input(input_string)
    rv = []
    while True:
      tok = lexer.token()
      if not tok:
        return rv
      rv.append(tok)

  def _SingleTokenForInput(self, input_string):
    """Gets the single token for the given input string. (Raises an exception if
    the input string does not result in exactly one token.)"""
    toks = self._TokensForInput(input_string)
    assert len(toks) == 1
    return toks[0]


if __name__ == "__main__":
  unittest.main()



# Copyright 2023 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import unittest

import mojom.generate.check as check
from mojom_bindings_generator import LoadChecks, _Generate
from mojom_parser_test_case import MojomParserTestCase


class FakeArgs:
  """Fakes args to _Generate - intention is to do just enough to run checks"""
  def __init__(self, tester, files=None):
    """ `tester` is MojomParserTestCase for paths.
        `files` will have tester path added."""
    self.checks_string = 'features'
    self.depth = tester.GetPath('')
    self.filelist = None
    self.filename = [tester.GetPath(x) for x in files]
    self.gen_directories = tester.GetPath('gen')
    self.generators_string = ''
    self.import_directories = []
    self.output_dir = tester.GetPath('out')
    self.scrambled_message_id_salt_paths = None
    self.typemaps = []
    self.variant = 'none'


class MojoBindingsCheckTest(MojomParserTestCase):
  def _ParseAndGenerate(self, mojoms):
    self.ParseMojoms(mojoms)
    args = FakeArgs(self, files=mojoms)
    _Generate(args, {})

  def assertValid(self, filename, content):
    self.WriteFile(filename, content)
    self._ParseAndGenerate([filename])

  def assertThrows(self, filename, content, regexp):
    mojoms = []
    self.WriteFile(filename, content)
    mojoms.append(filename)
    with self.assertRaisesRegexp(check.CheckException, regexp):
      self._ParseAndGenerate(mojoms)

  def testLoads(self):
    """Validate that the check is registered under the expected name."""
    check_modules = LoadChecks('features')
    self.assertTrue(check_modules['features'])

  def testNullableOk(self):
    self.assertValid(
        "a.mojom", """
          module a;
          // Scaffolding.
          feature kFeature {
            const string name = "Hello";
            const bool enabled_state = false;
          };
          [RuntimeFeature=kFeature]
          interface Guarded {
          };

          // Unguarded interfaces should be ok everywhere.
          interface NotGuarded { };

          // Optional (nullable) interfaces should be ok everywhere:
          struct Bar {
            pending_remote<Guarded>? remote;
            pending_receiver<Guarded>? receiver;
          };
          union Thingy {
            pending_remote<Guarded>? remote;
            pending_receiver<Guarded>? receiver;
          };
          interface Foo {
            Foo(
              pending_remote<Guarded>? remote,
              pending_receiver<Guarded>? receiver,
              pending_associated_remote<Guarded>? a_remote,
              pending_associated_receiver<Guarded>? a_receiver,
              // Unguarded interfaces do not have to be nullable.
              pending_remote<NotGuarded> remote,
              pending_receiver<NotGuarded> receiver,
              pending_associated_remote<NotGuarded> a_remote,
              pending_associated_receiver<NotGuarded> a_receiver
            ) => (
              pending_remote<Guarded>? remote,
              pending_receiver<Guarded>? receiver
            );
            Bar(array<pending_remote<Guarded>?> remote)
              => (map<string, pending_receiver<Guarded>?> a);
          };
    """)

  def testMethodParamsMustBeNullable(self):
    prelude = """
      module a;
      // Scaffolding.
      feature kFeature {
        const string name = "Hello";
        const bool enabled_state = false;
      };
      [RuntimeFeature=kFeature]
      interface Guarded { };
    """
    self.assertThrows(
        'a.mojom', prelude + """
          interface Trial {
            Method(pending_remote<Guarded> a) => ();
          };
                     """, 'interface Guarded has a RuntimeFeature')
    self.assertThrows(
        'a.mojom', prelude + """
          interface Trial {
            Method(bool foo) => (pending_receiver<Guarded> a);
          };
                     """, 'interface Guarded has a RuntimeFeature')
    self.assertThrows(
        'a.mojom', prelude + """
          interface Trial {
            Method(pending_receiver<Guarded> a) => ();
          };
                     """, 'interface Guarded has a RuntimeFeature')
    self.assertThrows(
        'a.mojom', prelude + """
          interface Trial {
            Method(pending_associated_remote<Guarded> a) => ();
          };
                     """, 'interface Guarded has a RuntimeFeature')
    self.assertThrows(
        'a.mojom', prelude + """
          interface Trial {
            Method(pending_associated_receiver<Guarded> a) => ();
          };
                     """, 'interface Guarded has a RuntimeFeature')
    self.assertThrows(
        'a.mojom', prelude + """
          interface Trial {
            Method(array<pending_associated_receiver<Guarded>> a) => ();
          };
                     """, 'interface Guarded has a RuntimeFeature')
    self.assertThrows(
        'a.mojom', prelude + """
          interface Trial {
            Method(map<string, pending_associated_receiver<Guarded>> a) => ();
          };
                     """, 'interface Guarded has a RuntimeFeature')

  def testStructUnionMembersMustBeNullable(self):
    prelude = """
      module a;
      // Scaffolding.
      feature kFeature {
        const string name = "Hello";
        const bool enabled_state = false;
      };
      [RuntimeFeature=kFeature]
      interface Guarded { };
    """
    self.assertThrows(
        'a.mojom', prelude + """
          struct Trial {
            pending_remote<Guarded> a;
          };
                     """, 'interface Guarded has a RuntimeFeature')
    self.assertThrows(
        'a.mojom', prelude + """
          union Trial {
            pending_remote<Guarded> a;
          };
                     """, 'interface Guarded has a RuntimeFeature')



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2022, Paul Elder <paul.elder@ideasonboard.com>

from libtuning.parsers.raspberrypi_parser import RaspberryPiParser
from libtuning.parsers.yaml_parser import YamlParser



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2022, Tomi Valkeinen <tomi.valkeinen@ideasonboard.com>

from collections import defaultdict
import gc
import libcamera as libcam
import selectors
import typing
import unittest
import weakref


class BaseTestCase(unittest.TestCase):
    def assertZero(self, a, msg=None):
        self.assertEqual(a, 0, msg)

    def assertIsAlive(self, wr, msg='object not alive'):
        self.assertIsNotNone(wr(), msg)

    def assertIsDead(self, wr, msg='object not dead'):
        self.assertIsNone(wr(), msg)

    def assertIsAllAlive(self, wr_list, msg='object not alive'):
        self.assertTrue(all([wr() for wr in wr_list]), msg)

    def assertIsAllDead(self, wr_list, msg='object not dead'):
        self.assertTrue(all([not wr() for wr in wr_list]), msg)


class SimpleTestMethods(BaseTestCase):
    def test_get_ref(self):
        cm = libcam.CameraManager.singleton()
        wr_cm = weakref.ref(cm)

        cam = cm.get('platform/vimc.0 Sensor B')
        self.assertIsNotNone(cam)
        wr_cam = weakref.ref(cam)

        del cm
        gc.collect()
        self.assertIsAlive(wr_cm)

        del cam
        gc.collect()
        self.assertIsDead(wr_cm)
        self.assertIsDead(wr_cam)

    def test_acquire_release(self):
        cm = libcam.CameraManager.singleton()
        cam = cm.get('platform/vimc.0 Sensor B')
        self.assertIsNotNone(cam)

        cam.acquire()

        cam.release()

    def test_double_acquire(self):
        cm = libcam.CameraManager.singleton()
        cam = cm.get('platform/vimc.0 Sensor B')
        self.assertIsNotNone(cam)

        cam.acquire()

        libcam.log_set_level('Camera', 'FATAL')
        with self.assertRaises(RuntimeError):
            cam.acquire()
        libcam.log_set_level('Camera', 'INFO')

        cam.release()

        # I expected exception here, but looks like double release works fine
        cam.release()

    def test_version(self):
        cm = libcam.CameraManager.singleton()
        self.assertIsInstance(cm.version, str)


class CameraTesterBase(BaseTestCase):
    cm: typing.Any
    cam: typing.Any

    def setUp(self):
        self.cm = libcam.CameraManager.singleton()
        self.cam = next((cam for cam in self.cm.cameras if 'platform/vimc' in cam.id), None)
        if self.cam is None:
            self.cm = None
            self.skipTest('No vimc found')

        self.cam.acquire()

        self.wr_cam = weakref.ref(self.cam)
        self.wr_cm = weakref.ref(self.cm)

    def tearDown(self):
        # If a test fails, the camera may be in running state. So always stop.
        self.cam.stop()

        self.cam.release()

        self.cam = None
        self.cm = None

        self.assertIsDead(self.wr_cm)
        self.assertIsDead(self.wr_cam)


class AllocatorTestMethods(CameraTesterBase):
    def test_allocator(self):
        cam = self.cam

        camconfig = cam.generate_configuration([libcam.StreamRole.StillCapture])
        self.assertTrue(camconfig.size == 1)
        wr_camconfig = weakref.ref(camconfig)

        streamconfig = camconfig.at(0)
        wr_streamconfig = weakref.ref(streamconfig)

        cam.configure(camconfig)

        stream = streamconfig.stream
        wr_stream = weakref.ref(stream)

        # stream should keep streamconfig and camconfig alive
        del streamconfig
        del camconfig
        gc.collect()
        self.assertIsAlive(wr_camconfig)
        self.assertIsAlive(wr_streamconfig)

        allocator = libcam.FrameBufferAllocator(cam)
        num_bufs = allocator.allocate(stream)
        self.assertTrue(num_bufs > 0)
        wr_allocator = weakref.ref(allocator)

        buffers = allocator.buffers(stream)
        self.assertIsNotNone(buffers)
        del buffers

        buffer = allocator.buffers(stream)[0]
        self.assertIsNotNone(buffer)
        wr_buffer = weakref.ref(buffer)

        del allocator
        gc.collect()
        self.assertIsAlive(wr_buffer)
        self.assertIsAlive(wr_allocator)
        self.assertIsAlive(wr_stream)

        del buffer
        gc.collect()
        self.assertIsDead(wr_buffer)
        self.assertIsDead(wr_allocator)
        self.assertIsAlive(wr_stream)

        del stream
        gc.collect()
        self.assertIsDead(wr_stream)
        self.assertIsDead(wr_camconfig)
        self.assertIsDead(wr_streamconfig)


class SimpleCaptureMethods(CameraTesterBase):
    def test_blocking(self):
        cm = self.cm
        cam = self.cam

        camconfig = cam.generate_configuration([libcam.StreamRole.StillCapture])
        self.assertTrue(camconfig.size == 1)

        streamconfig = camconfig.at(0)
        fmts = streamconfig.formats
        self.assertIsNotNone(fmts)
        fmts = None

        cam.configure(camconfig)

        stream = streamconfig.stream

        allocator = libcam.FrameBufferAllocator(cam)
        num_bufs = allocator.allocate(stream)
        self.assertTrue(num_bufs > 0)

        num_bufs = len(allocator.buffers(stream))

        reqs = []
        for i in range(num_bufs):
            req = cam.create_request(i)
            self.assertIsNotNone(req)

            buffer = allocator.buffers(stream)[i]
            req.add_buffer(stream, buffer)

            reqs.append(req)

        buffer = None

        cam.start()

        for req in reqs:
            cam.queue_request(req)

        reqs = None
        gc.collect()

        sel = selectors.DefaultSelector()
        sel.register(cm.event_fd, selectors.EVENT_READ)

        reqs = []

        while True:
            events = sel.select()
            if not events:
                continue

            ready_reqs = cm.get_ready_requests()

            reqs += ready_reqs

            if len(reqs) == num_bufs:
                break

        for i, req in enumerate(reqs):
            self.assertTrue(i == req.cookie)

        reqs = None
        gc.collect()

        cam.stop()

    def test_select(self):
        cm = self.cm
        cam = self.cam

        camconfig = cam.generate_configuration([libcam.StreamRole.StillCapture])
        self.assertTrue(camconfig.size == 1)

        streamconfig = camconfig.at(0)
        fmts = streamconfig.formats
        self.assertIsNotNone(fmts)
        fmts = None

        cam.configure(camconfig)

        stream = streamconfig.stream

        allocator = libcam.FrameBufferAllocator(cam)
        num_bufs = allocator.allocate(stream)
        self.assertTrue(num_bufs > 0)

        num_bufs = len(allocator.buffers(stream))

        reqs = []
        for i in range(num_bufs):
            req = cam.create_request(i)
            self.assertIsNotNone(req)

            buffer = allocator.buffers(stream)[i]
            req.add_buffer(stream, buffer)

            reqs.append(req)

        buffer = None

        cam.start()

        for req in reqs:
            cam.queue_request(req)

        reqs = None
        gc.collect()

        sel = selectors.DefaultSelector()
        sel.register(cm.event_fd, selectors.EVENT_READ)

        reqs = []

        running = True
        while running:
            events = sel.select()
            for _ in events:
                ready_reqs = cm.get_ready_requests()

                reqs += ready_reqs

                if len(reqs) == num_bufs:
                    running = False

        self.assertTrue(len(reqs) == num_bufs)

        for i, req in enumerate(reqs):
            self.assertTrue(i == req.cookie)

        reqs = None
        gc.collect()

        cam.stop()


# Recursively expand slist's objects into olist, using seen to track already
# processed objects.
def _getr(slist, olist, seen):
    for e in slist:
        if id(e) in seen:
            continue
        seen.add(id(e))
        olist.append(e)
        tl = gc.get_referents(e)
        if tl:
            _getr(tl, olist, seen)


def get_all_objects(ignored=[]):
    gcl = gc.get_objects()
    olist = []
    seen = set()

    seen.add(id(gcl))
    seen.add(id(olist))
    seen.add(id(seen))
    seen.update(set([id(o) for o in ignored]))

    _getr(gcl, olist, seen)

    return olist


def create_type_count_map(olist):
    map = defaultdict(int)
    for o in olist:
        map[type(o)] += 1
    return map


def diff_type_count_maps(before, after):
    return [(k, after[k] - before[k]) for k in after if after[k] != before[k]]


if __name__ == '__main__':
    # \todo This is an attempt to see the Python objects that are not collected,
    # but this doesn't work very well, as things always leak a bit.
    test_leaks = False

    if test_leaks:
        gc.unfreeze()
        gc.collect()

        obs_before = get_all_objects()

    unittest.main(exit=False)

    if test_leaks:
        gc.unfreeze()
        gc.collect()

        obs_after = get_all_objects([obs_before])   # type: ignore

        before = create_type_count_map(obs_before)  # type: ignore
        after = create_type_count_map(obs_after)

        leaks = diff_type_count_maps(before, after)
        if len(leaks) > 0:
            print(leaks)



# Copyright 2020 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from mojom_parser_test_case import MojomParserTestCase
from mojom.generate import module as mojom


class ConstTest(MojomParserTestCase):
  """Tests constant parsing behavior."""

  def testLiteralInt(self):
    a_mojom = 'a.mojom'
    self.WriteFile(a_mojom, 'const int32 k = 42;')
    self.ParseMojoms([a_mojom])
    a = self.LoadModule(a_mojom)
    self.assertEqual(1, len(a.constants))
    self.assertEqual('k', a.constants[0].mojom_name)
    self.assertEqual('42', a.constants[0].value)

  def testLiteralFloat(self):
    a_mojom = 'a.mojom'
    self.WriteFile(a_mojom, 'const float k = 42.5;')
    self.ParseMojoms([a_mojom])
    a = self.LoadModule(a_mojom)
    self.assertEqual(1, len(a.constants))
    self.assertEqual('k', a.constants[0].mojom_name)
    self.assertEqual('42.5', a.constants[0].value)

  def testLiteralString(self):
    a_mojom = 'a.mojom'
    self.WriteFile(a_mojom, 'const string k = "woot";')
    self.ParseMojoms([a_mojom])
    a = self.LoadModule(a_mojom)
    self.assertEqual(1, len(a.constants))
    self.assertEqual('k', a.constants[0].mojom_name)
    self.assertEqual('"woot"', a.constants[0].value)

  def testEnumConstant(self):
    a_mojom = 'a.mojom'
    self.WriteFile(a_mojom, 'module a; enum E { kA = 41, kB };')
    b_mojom = 'b.mojom'
    self.WriteFile(
        b_mojom, """\
      import "a.mojom";
      const a.E kE1 = a.E.kB;

      // We also allow value names to be unqualified, implying scope from the
      // constant's type.
      const a.E kE2 = kB;
      """)
    self.ParseMojoms([a_mojom, b_mojom])
    a = self.LoadModule(a_mojom)
    b = self.LoadModule(b_mojom)
    self.assertEqual(1, len(a.enums))
    self.assertEqual('E', a.enums[0].mojom_name)
    self.assertEqual(2, len(b.constants))
    self.assertEqual('kE1', b.constants[0].mojom_name)
    self.assertEqual(a.enums[0], b.constants[0].kind)
    self.assertEqual(a.enums[0].fields[1], b.constants[0].value.field)
    self.assertEqual(42, b.constants[0].value.field.numeric_value)
    self.assertEqual('kE2', b.constants[1].mojom_name)
    self.assertEqual(a.enums[0].fields[1], b.constants[1].value.field)
    self.assertEqual(42, b.constants[1].value.field.numeric_value)

  def testConstantReference(self):
    a_mojom = 'a.mojom'
    self.WriteFile(a_mojom, 'const int32 kA = 42; const int32 kB = kA;')
    self.ParseMojoms([a_mojom])
    a = self.LoadModule(a_mojom)
    self.assertEqual(2, len(a.constants))
    self.assertEqual('kA', a.constants[0].mojom_name)
    self.assertEqual('42', a.constants[0].value)
    self.assertEqual('kB', a.constants[1].mojom_name)
    self.assertEqual('42', a.constants[1].value)

  def testImportedConstantReference(self):
    a_mojom = 'a.mojom'
    self.WriteFile(a_mojom, 'const int32 kA = 42;')
    b_mojom = 'b.mojom'
    self.WriteFile(b_mojom, 'import "a.mojom"; const int32 kB = kA;')
    self.ParseMojoms([a_mojom, b_mojom])
    a = self.LoadModule(a_mojom)
    b = self.LoadModule(b_mojom)
    self.assertEqual(1, len(a.constants))
    self.assertEqual(1, len(b.constants))
    self.assertEqual('kA', a.constants[0].mojom_name)
    self.assertEqual('42', a.constants[0].value)
    self.assertEqual('kB', b.constants[0].mojom_name)
    self.assertEqual('42', b.constants[0].value)



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
# Copyright (C) 2022, Paul Elder <paul.elder@ideasonboard.com>
#
# Utilities for libtuning

import cv2
import decimal
import math
import numpy as np
import os
from pathlib import Path
import re
import sys
import logging

import libtuning as lt
from libtuning.image import Image
from .macbeth import locate_macbeth

logger = logging.getLogger(__name__)

# Utility functions


def get_module_by_type_name(modules, name):
    for module in modules:
        if module.type == name:
            return module
    return None


# Private utility functions


def _list_image_files(directory):
    d = Path(directory)
    files = [d.joinpath(f) for f in os.listdir(d)
             if re.search(r'\.(jp[e]g$)|(dng$)', f)]
    files.sort()
    return files


def _parse_image_filename(fn: Path):
    lsc_only = False
    color_temperature = None
    lux = None

    parts = fn.stem.split('_')
    for part in parts:
        if part == 'alsc':
            lsc_only = True
            continue
        r = re.match(r'(\d+)[kK]', part)
        if r:
            color_temperature = int(r.group(1))
            continue
        r = re.match(r'(\d+)[lLuU]', part)
        if r:
            lux = int(r.group(1))

    if color_temperature is None:
        logger.error(f'The file name of "{fn.name}" does not contain a color temperature')

    if lux is None and lsc_only is False:
        logger.error(f'The file name of "{fn.name}" must either contain alsc or a lux level')

    return color_temperature, lux, lsc_only


# \todo Implement this from check_imgs() in ctt.py
def _validate_images(images):
    return True


# Public utility functions


# @brief Load images into a single list of Image instances
# @param input_dir Directory from which to load image files
# @param config Configuration dictionary
# @param load_nonlsc Whether or not to load non-lsc images
# @param load_lsc Whether or not to load lsc-only images
# @return A list of Image instances
def load_images(input_dir: str, config: dict, load_nonlsc: bool, load_lsc: bool) -> list:
    files = _list_image_files(input_dir)
    if len(files) == 0:
        logger.error(f'No images found in {input_dir}')
        return None

    images = []
    for f in files:
        color, lux, lsc_only = _parse_image_filename(f)

        if color is None:
            logger.warning(f'Ignoring "{f.name}" as it has no associated color temperature')
            continue

        logger.info(f'Process image "{f.name}" (color={color}, lux={lux}, lsc_only={lsc_only})')

        # Skip lsc image if we don't need it
        if lsc_only and not load_lsc:
            logger.warning(f'Skipping {f.name} as this tuner has no LSC module')
            continue

        # Skip non-lsc image if we don't need it
        if not lsc_only and not load_nonlsc:
            logger.warning(f'Skipping {f.name} as this tuner only has an LSC module')
            continue

        # Load image
        try:
            image = Image(f)
        except Exception as e:
            logger.error(f'Failed to load image {f.name}: {e}')
            continue

        # Populate simple fields
        image.lsc_only = lsc_only
        image.color = color
        image.lux = lux

        # Black level comes from the TIFF tags, but they are overridable by the
        # config file.
        if 'blacklevel' in config['general']:
            image.blacklevel_16 = config['general']['blacklevel']

        if lsc_only:
            images.append(image)
            continue

        # Handle macbeth
        macbeth = locate_macbeth(image, config)
        if macbeth is None:
            continue

        images.append(image)

    if not _validate_images(images):
        return None

    return images



"""
Some code that will save virtual macbeth charts that show the difference between optimised matrices and non optimised matrices

The function creates an image that is 1550 by 1050 pixels wide, and fills it with patches which are 200x200 pixels in size
Each patch contains the ideal color, the color from the original matrix, and the color from the final matrix
_________________
|               |
|  Ideal Color  |
|_______________|
|  Old  |  new  |
| Color | Color |
|_______|_______|

Nice way of showing how the optimisation helps change the colors and the color matricies
"""
def visualise_macbeth_chart(macbeth_rgb, original_rgb, new_rgb, output_filename):
    image = np.zeros((1050, 1550, 3), dtype=np.uint8)
    colorindex = -1
    for y in range(6):
        for x in range(4):  # Creates 6 x 4 grid of macbeth chart
            colorindex += 1
            xlocation = 50 + 250 * x  # Means there is 50px of black gap between each square, more like the real macbeth chart.
            ylocation = 50 + 250 * y
            for g in range(200):
                for i in range(100):
                    image[xlocation + i, ylocation + g] = macbeth_rgb[colorindex]
            xlocation = 150 + 250 * x
            ylocation = 50 + 250 * y
            for i in range(100):
                for g in range(100):
                    image[xlocation + i, ylocation + g] = original_rgb[colorindex]  # Smaller squares below to compare the old colors with the new ones
            xlocation = 150 + 250 * x
            ylocation = 150 + 250 * y
            for i in range(100):
                for g in range(100):
                    image[xlocation + i, ylocation + g] = new_rgb[colorindex]

    im_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    cv2.imwrite(f'{output_filename} Generated Macbeth Chart.png', im_bgr)




#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2019, Google Inc.
#
# Author: Laurent Pinchart <laurent.pinchart@ideasonboard.com>
#
# Generate control definitions from YAML

import argparse
import jinja2
import os
import sys
import yaml

from controls import Control


def snake_case(s):
    return ''.join([c.isupper() and ('_' + c) or c for c in s]).strip('_')


def format_description(description):
    description = description.strip('\n').split('\n')
    for i in range(1, len(description)):
        line = description[i]
        description[i] = (line and ' * ' or ' *') + line
    return '\n'.join(description)


def extend_control(ctrl, id, ranges):
    ctrl.id = ranges[ctrl.vendor] + id + 1

    if ctrl.vendor != 'libcamera':
        ctrl.namespace = f'{ctrl.vendor}::'
    else:
        ctrl.namespace = ''

    return ctrl


def main(argv):

    # Parse command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--mode', '-m', type=str, required=True, choices=['controls', 'properties'],
                        help='Mode of operation')
    parser.add_argument('--output', '-o', metavar='file', type=str,
                        help='Output file name. Defaults to standard output if not specified.')
    parser.add_argument('--ranges', '-r', type=str, required=True,
                        help='Control id range reservation file.')
    parser.add_argument('--template', '-t', dest='template', type=str, required=True,
                        help='Template file name.')
    parser.add_argument('input', type=str, nargs='+',
                        help='Input file name.')

    args = parser.parse_args(argv[1:])

    ranges = {}
    with open(args.ranges, 'rb') as f:
        data = open(args.ranges, 'rb').read()
        ranges = yaml.safe_load(data)['ranges']

    controls = {}
    for input in args.input:
        data = yaml.safe_load(open(input, 'rb').read())

        vendor = data['vendor']
        if vendor not in ranges.keys():
            raise RuntimeError(f'Control id range is not defined for vendor {vendor}')

        ctrls = controls.setdefault(vendor, [])

        for i, ctrl in enumerate(data['controls']):
            ctrl = Control(*ctrl.popitem(), vendor, args.mode)
            ctrls.append(extend_control(ctrl, i, ranges))

    # Sort the vendors by range numerical value
    controls = [[vendor, ctrls] for vendor, ctrls in controls.items()]
    controls.sort(key=lambda item: ranges[item[0]])

    filename = {
        'controls': 'control_ids',
        'properties': 'property_ids',
    }[args.mode]

    data = {
        'filename': filename,
        'mode': args.mode,
        'controls': controls,
    }

    env = jinja2.Environment()
    env.filters['format_description'] = format_description
    env.filters['snake_case'] = snake_case
    template = env.from_string(open(args.template, 'r', encoding='utf-8').read())
    string = template.render(data)

    if args.output:
        output = open(args.output, 'w', encoding='utf-8')
        output.write(string)
        output.close()
    else:
        sys.stdout.write(string)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))



# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2022, Tomi Valkeinen <tomi.valkeinen@ideasonboard.com>

import selectors
import sys


class NullRenderer:
    def __init__(self, state):
        self.state = state

        self.cm = state.cm
        self.contexts = state.contexts

        self.running = False

    def setup(self):
        pass

    def run(self):
        print('Capturing...')

        self.running = True

        sel = selectors.DefaultSelector()
        sel.register(self.cm.event_fd, selectors.EVENT_READ, self.readcam)
        sel.register(sys.stdin, selectors.EVENT_READ, self.readkey)

        print('Press enter to exit')

        while self.running:
            events = sel.select()
            for key, mask in events:
                callback = key.data
                callback(key.fileobj)

        print('Exiting...')

    def readcam(self, fd):
        self.running = self.state.event_handler()

    def readkey(self, fileobj):
        sys.stdin.readline()
        self.running = False

    def request_handler(self, ctx, req):
        self.state.request_processed(ctx, req)



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2024, Paul Elder <paul.elder@ideasonboard.com>

from libtuning.modules.agc.agc import AGC
from libtuning.modules.agc.rkisp1 import AGCRkISP1



# Copyright 2014 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import unittest

from mojom_bindings_generator import MakeImportStackMessage
from mojom_bindings_generator import ScrambleMethodOrdinals


class FakeIface:
  def __init__(self):
    self.mojom_name = None
    self.methods = None


class FakeMethod:
  def __init__(self, explicit_ordinal=None):
    self.explicit_ordinal = explicit_ordinal
    self.ordinal = explicit_ordinal
    self.ordinal_comment = None


class MojoBindingsGeneratorTest(unittest.TestCase):
  """Tests mojo_bindings_generator."""

  def testMakeImportStackMessage(self):
    """Tests MakeImportStackMessage()."""
    self.assertEqual(MakeImportStackMessage(["x"]), "")
    self.assertEqual(MakeImportStackMessage(["x", "y"]),
                     "\n  y was imported by x")
    self.assertEqual(MakeImportStackMessage(["x", "y", "z"]),
                     "\n  z was imported by y\n  y was imported by x")

  def testScrambleMethodOrdinals(self):
    """Tests ScrambleMethodOrdinals()."""
    interface = FakeIface()
    interface.mojom_name = 'RendererConfiguration'
    interface.methods = [
        FakeMethod(),
        FakeMethod(),
        FakeMethod(),
        FakeMethod(explicit_ordinal=42)
    ]
    ScrambleMethodOrdinals([interface], "foo".encode('utf-8'))
    # These next three values are hard-coded. If the generation algorithm
    # changes from being based on sha256(seed + interface.name + str(i)) then
    # these numbers will obviously need to change too.
    #
    # Note that hashlib.sha256('fooRendererConfiguration1').digest()[:4] is
    # '\xa5\xbc\xf9\xca' and that hex(1257880741) = '0x4af9bca5'. The
    # difference in 0x4a vs 0xca is because we only take 31 bits.
    self.assertEqual(interface.methods[0].ordinal, 1257880741)
    self.assertEqual(interface.methods[1].ordinal, 631133653)
    self.assertEqual(interface.methods[2].ordinal, 549336076)

    # Explicit method ordinals should not be scrambled.
    self.assertEqual(interface.methods[3].ordinal, 42)


if __name__ == "__main__":
  unittest.main()



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2022, Paul Elder <paul.elder@ideasonboard.com>
#
# Parser for Raspberry Pi config file format

from .parser import Parser

import json
import numbers

import libtuning.utils as utils


class RaspberryPiParser(Parser):
    def __init__(self):
        super().__init__()

    # The string in the 'disable' and 'plot' lists are formatted as
    # 'rpi.{module_name}'.
    # @brief Enumerate, as a module, @a listt if its value exists in @a dictt
    #        and it is the name of a valid module in @a modules
    def _enumerate_rpi_modules(self, listt, dictt, modules):
        for x in listt:
            name = x.replace('rpi.', '')
            if name not in dictt:
                continue
            module = utils.get_module_by_typename(modules, name)
            if module is not None:
                yield module

    def _valid_macbeth_option(self, value):
        if not isinstance(value, dict):
            return False

        if list(value.keys()) != ['small', 'show']:
            return False

        for val in value.values():
            if not isinstance(val, numbers.Number):
                return False

        return True

    def parse(self, config_file: str, modules: list) -> (dict, list):
        with open(config_file, 'r') as config_json:
            config = json.load(config_json)

        disable = []
        for module in self._enumerate_rpi_modules(config['disable'], config, modules):
            disable.append(module)
            # Remove the disabled module's config too
            config.pop(module.name)
        config.pop('disable')

        # The raspberrypi config format has 'plot' map to a list of module
        # names which should be plotted. libtuning has each module contain the
        # plot information in itself so do this conversion.

        for module in self._enumerate_rpi_modules(config['plot'], config, modules):
            # It's fine to set the value of a potentially disabled module, as
            # the object still exists at this point
            module.appendValue('debug', 'plot')
        config.pop('plot')

        # Convert the keys from module name to module instance

        new_config = {}

        for module_name in config:
            module = utils.get_module_by_type_name(modules, module_name)
            if module is not None:
                new_config[module] = config[module_name]

        new_config['general'] = {}

        if 'blacklevel' in config:
            if not isinstance(config['blacklevel'], numbers.Number):
                raise TypeError('Config "blacklevel" must be a number')
            # Raspberry Pi's ctt config has magic blacklevel value -1 to mean
            # "get it from the image metadata". Since we already do that in
            # Image, don't save it to the config here.
            if config['blacklevel'] >= 0:
                new_config['general']['blacklevel'] = config['blacklevel']

        if 'macbeth' in config:
            if not self._valid_macbeth_option(config['macbeth']):
                raise TypeError('Config "macbeth" must be a dict: {"small": number, "show": number}')
            new_config['general']['macbeth'] = config['macbeth']
        else:
            new_config['general']['macbeth'] = {'small': 0, 'show': 0}

        return new_config, disable



# Copyright 2014 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.


class Error(Exception):
  """Base class for Mojo IDL bindings parser/generator errors."""

  def __init__(self, filename, message, lineno=None, addenda=None, **kwargs):
    """|filename| is the (primary) file which caused the error, |message| is the
    error message, |lineno| is the 1-based line number (or |None| if not
    applicable/available), and |addenda| is a list of additional lines to append
    to the final error message."""
    Exception.__init__(self, **kwargs)
    self.filename = filename
    self.message = message
    self.lineno = lineno
    self.addenda = addenda

  def __str__(self):
    if self.lineno:
      s = "%s:%d: Error: %s" % (self.filename, self.lineno, self.message)
    else:
      s = "%s: Error: %s" % (self.filename, self.message)
    return "\n".join([s] + self.addenda) if self.addenda else s

  def __repr__(self):
    return str(self)



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (C) 2022, Tomi Valkeinen <tomi.valkeinen@ideasonboard.com>

# A simple capture example showing:
# - How to setup the camera
# - Capture certain number of frames in a blocking manner
# - How to stop the camera
#
# This simple example is, in many ways, too simple. The purpose of the example
# is to introduce the concepts. A more realistic example is given in
# simple-continuous-capture.py.

import argparse
import libcamera as libcam
import selectors
import sys

# Number of frames to capture
TOTAL_FRAMES = 30


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--camera', type=str, default='1',
                        help='Camera index number (starting from 1) or part of the name')
    parser.add_argument('-f', '--format', type=str, help='Pixel format')
    parser.add_argument('-s', '--size', type=str, help='Size ("WxH")')
    args = parser.parse_args()

    cm = libcam.CameraManager.singleton()

    try:
        if args.camera.isnumeric():
            cam_idx = int(args.camera)
            cam = next((cam for i, cam in enumerate(cm.cameras) if i + 1 == cam_idx))
        else:
            cam = next((cam for cam in cm.cameras if args.camera in cam.id))
    except Exception:
        print(f'Failed to find camera "{args.camera}"')
        return -1

    # Acquire the camera for our use

    cam.acquire()

    # Configure the camera

    cam_config = cam.generate_configuration([libcam.StreamRole.Viewfinder])

    stream_config = cam_config.at(0)

    if args.format:
        fmt = libcam.PixelFormat(args.format)
        stream_config.pixel_format = fmt

    if args.size:
        w, h = [int(v) for v in args.size.split('x')]
        stream_config.size = libcam.Size(w, h)

    cam.configure(cam_config)

    print(f'Capturing {TOTAL_FRAMES} frames with {stream_config}')

    stream = stream_config.stream

    # Allocate the buffers for capture

    allocator = libcam.FrameBufferAllocator(cam)
    ret = allocator.allocate(stream)
    assert ret > 0

    num_bufs = len(allocator.buffers(stream))

    # Create the requests and assign a buffer for each request

    reqs = []
    for i in range(num_bufs):
        # Use the buffer index as the cookie
        req = cam.create_request(i)

        buffer = allocator.buffers(stream)[i]
        req.add_buffer(stream, buffer)

        reqs.append(req)

    # Start the camera

    cam.start()

    # frames_queued and frames_done track the number of frames queued and done

    frames_queued = 0
    frames_done = 0

    # Queue the requests to the camera

    for req in reqs:
        cam.queue_request(req)
        frames_queued += 1

    # The main loop. Wait for the queued Requests to complete, process them,
    # and re-queue them again.

    sel = selectors.DefaultSelector()
    sel.register(cm.event_fd, selectors.EVENT_READ)

    while frames_done < TOTAL_FRAMES:
        # cm.get_ready_requests() does not block, so we use a Selector to wait
        # for a camera event. Here we should almost always get a single
        # Request, but in some cases there could be multiple or none.

        events = sel.select()
        if not events:
            continue

        reqs = cm.get_ready_requests()

        for req in reqs:
            frames_done += 1

            buffers = req.buffers

            # A ready Request could contain multiple buffers if multiple streams
            # were being used. Here we know we only have a single stream,
            # and we use next(iter()) to get the first and only buffer.

            assert len(buffers) == 1

            stream, fb = next(iter(buffers.items()))

            # Here we could process the received buffer. In this example we only
            # print a few details below.

            meta = fb.metadata

            print("seq {:3}, bytes {}, frames queued/done {:3}/{:<3}"
                  .format(meta.sequence,
                          '/'.join([str(p.bytes_used) for p in meta.planes]),
                          frames_queued, frames_done))

            # If we want to capture more frames we need to queue more Requests.
            # We could create a totally new Request, but it is more efficient
            # to reuse the existing one that we just received.
            if frames_queued < TOTAL_FRAMES:
                req.reuse()
                cam.queue_request(req)
                frames_queued += 1

    # Stop the camera

    cam.stop()

    # Release the camera

    cam.release()

    return 0


if __name__ == '__main__':
    sys.exit(main())



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
#
# Container for an image and associated metadata

import binascii
import numpy as np
from pathlib import Path
import pyexiv2 as pyexif
import rawpy as raw
import re

import libtuning as lt
import libtuning.utils as utils
import logging

logger = logging.getLogger(__name__)


class Image:
    def __init__(self, path: Path):
        self.path = path
        self.lsc_only = False
        self.color = -1
        self.lux = -1
        self.macbeth = None

        try:
            self._load_metadata_exif()
        except Exception as e:
            logger.error(f'Failed to load metadata from {self.path}: {e}')
            raise e

        try:
            self._read_image_dng()
        except Exception as e:
            logger.error(f'Failed to load image data from {self.path}: {e}')
            raise e

    @property
    def name(self):
        return self.path.name

    # May raise KeyError as there are too many to check
    def _load_metadata_exif(self):
        # RawPy doesn't load all the image tags that we need, so we use py3exiv2
        metadata = pyexif.ImageMetadata(str(self.path))
        metadata.read()

        # The DNG and TIFF/EP specifications use different IFDs to store the
        # raw image data and the Exif tags. DNG stores them in a SubIFD and in
        # an Exif IFD respectively (named "SubImage1" and "Photo" by pyexiv2),
        # while TIFF/EP stores them both in IFD0 (name "Image"). Both are used
        # in "DNG" files, with libcamera-apps following the DNG recommendation
        # and applications based on picamera2 following TIFF/EP.
        #
        # This code detects which tags are being used, and therefore extracts the
        # correct values.
        try:
            self.w = metadata['Exif.SubImage1.ImageWidth'].value
            subimage = 'SubImage1'
            photo = 'Photo'
        except KeyError:
            self.w = metadata['Exif.Image.ImageWidth'].value
            subimage = 'Image'
            photo = 'Image'
        self.pad = 0
        self.h = metadata[f'Exif.{subimage}.ImageLength'].value
        white = metadata[f'Exif.{subimage}.WhiteLevel'].value
        self.sigbits = int(white).bit_length()
        self.fmt = (self.sigbits - 4) // 2
        self.exposure = int(metadata[f'Exif.{photo}.ExposureTime'].value * 1000000)
        self.againQ8 = metadata[f'Exif.{photo}.ISOSpeedRatings'].value * 256 / 100
        self.againQ8_norm = self.againQ8 / 256
        self.camName = metadata['Exif.Image.Model'].value
        self.blacklevel = int(metadata[f'Exif.{subimage}.BlackLevel'].value[0])
        self.blacklevel_16 = self.blacklevel << (16 - self.sigbits)

        # Channel order depending on bayer pattern
        # The key is the order given by exif, where 0 is R, 1 is G, and 2 is B
        # The value is the index where the color can be found, where the first
        # is R, then G, then G, then B.
        bayer_case = {
            '0 1 1 2': (lt.Color.R, lt.Color.GR, lt.Color.GB, lt.Color.B),
            '1 2 0 1': (lt.Color.GB, lt.Color.B, lt.Color.R, lt.Color.GR),
            '2 1 1 0': (lt.Color.B, lt.Color.GB, lt.Color.GR, lt.Color.R),
            '1 0 2 1': (lt.Color.GR, lt.Color.R, lt.Color.B, lt.Color.GB)
        }
        # Note: This needs to be in IFD0
        cfa_pattern = metadata[f'Exif.{subimage}.CFAPattern'].value
        self.order = bayer_case[cfa_pattern]

    def _read_image_dng(self):
        raw_im = raw.imread(str(self.path))
        raw_data = raw_im.raw_image
        shift = 16 - self.sigbits
        c0 = np.left_shift(raw_data[0::2, 0::2].astype(np.int64), shift)
        c1 = np.left_shift(raw_data[0::2, 1::2].astype(np.int64), shift)
        c2 = np.left_shift(raw_data[1::2, 0::2].astype(np.int64), shift)
        c3 = np.left_shift(raw_data[1::2, 1::2].astype(np.int64), shift)
        self.channels = [c0, c1, c2, c3]
        # Reorder the channels into R, GR, GB, B
        self.channels = [self.channels[i] for i in self.order]

    # \todo Move this to macbeth.py
    def get_patches(self, cen_coords, size=16):
        saturated = False

        # Obtain channel widths and heights
        ch_w, ch_h = self.w, self.h
        cen_coords = list(np.array((cen_coords[0])).astype(np.int32))
        self.cen_coords = cen_coords

        # Squares are ordered by stacking macbeth chart columns from left to
        # right. Some useful patch indices:
        #     white = 3
        #     black = 23
        #     'reds' = 9, 10
        #     'blues' = 2, 5, 8, 20, 22
        #     'greens' = 6, 12, 17
        #     greyscale = 3, 7, 11, 15, 19, 23
        all_patches = []
        for ch in self.channels:
            ch_patches = []
            for cen in cen_coords:
                # Macbeth centre is placed at top left of central 2x2 patch to
                # account for rounding. Patch pixels are sorted by pixel
                # brightness so spatial information is lost.
                patch = ch[cen[1] - 7:cen[1] + 9, cen[0] - 7:cen[0] + 9].flatten()
                patch.sort()
                if patch[-5] == (2**self.sigbits - 1) * 2**(16 - self.sigbits):
                    saturated = True
                ch_patches.append(patch)

            all_patches.append(ch_patches)

        self.patches = np.array(all_patches)

        return not saturated



# Copyright 2022 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Code shared by the various pre-generation mojom checkers."""


class CheckException(Exception):
  def __init__(self, module, message):
    self.module = module
    self.message = message
    super().__init__(self.message)

  def __str__(self):
    return "Failed mojo pre-generation check for {}:\n{}".format(
        self.module.path, self.message)


class Check:
  def __init__(self, module):
    self.module = module

  def CheckModule(self):
    """ Subclass should return True if its Checks pass, and throw an
    exception otherwise. CheckModule will be called immediately before
    mojom.generate.Generator.GenerateFiles()"""
    raise NotImplementedError("Subclasses must override/implement this method")



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2022, Paul Elder <paul.elder@ideasonboard.com>
# Copyright (C) 2024, Ideas On Board
#
# Tuning script for rkisp1

import coloredlogs
import logging
import sys

import libtuning as lt
from libtuning.parsers import YamlParser
from libtuning.generators import YamlOutput
from libtuning.modules.lsc import LSCRkISP1
from libtuning.modules.agc import AGCRkISP1
from libtuning.modules.awb import AWBRkISP1
from libtuning.modules.ccm import CCMRkISP1
from libtuning.modules.static import StaticModule

coloredlogs.install(level=logging.INFO, fmt='%(name)s %(levelname)s %(message)s')

agc = AGCRkISP1(debug=[lt.Debug.Plot])
awb = AWBRkISP1(debug=[lt.Debug.Plot])
blc = StaticModule('BlackLevelCorrection')
ccm = CCMRkISP1(debug=[lt.Debug.Plot])
color_processing = StaticModule('ColorProcessing')
filter = StaticModule('Filter')
gamma_out = StaticModule('GammaOutCorrection', {'gamma': 2.2})
lsc = LSCRkISP1(debug=[lt.Debug.Plot],
                # This is for the actual LSC tuning, and is part of the base LSC
                # module. rkisp1's table sector sizes (16x16 programmed as mirrored
                # 8x8) are separate, and is hardcoded in its specific LSC tuning
                # module.
                sector_shape=(17, 17),

                sector_x_gradient=lt.gradient.Linear(lt.Remainder.DistributeFront),
                sector_y_gradient=lt.gradient.Linear(lt.Remainder.DistributeFront),

                # This is the function that will be used to average the pixels in
                # each sector. This can also be a custom function.
                sector_average_function=lt.average.Mean(),

                # This is the function that will be used to smooth the color ratio
                # values.  This can also be a custom function.
                smoothing_function=lt.smoothing.MedianBlur(3),)

tuner = lt.Tuner('RkISP1')
tuner.add([agc, awb, blc, ccm, color_processing, filter, gamma_out, lsc])
tuner.set_input_parser(YamlParser())
tuner.set_output_formatter(YamlOutput())
tuner.set_output_order([agc, awb, blc, ccm, color_processing,
                        filter, gamma_out, lsc])

if __name__ == '__main__':
    sys.exit(tuner.run(sys.argv))



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2020, Google Inc.
#
# Author: Paul Elder <paul.elder@ideasonboard.com>
#
# Generate header file to contain lttng tracepoints

import jinja2
import pathlib
import os
import sys

def main(argv):
    if len(argv) < 4:
        print(f'Usage: {argv[0]} include_build_dir output template tp_files...')
        return 1

    output = argv[2]
    template = argv[3]

    path = pathlib.Path(output).absolute().relative_to(argv[1])

    source = ''
    for fname in argv[4:]:
        source += open(fname, 'r', encoding='utf-8').read() + '\n\n'

    template = jinja2.Template(open(template, 'r', encoding='utf-8').read())
    string = template.render(path=path, source=source)

    f = open(output, 'w', encoding='utf-8').write(string)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2024, Paul Elder <paul.elder@ideasonboard.com>
# Copyright (C) 2024, Ideas on Board
#
# Ccm module for tuning rkisp1

from .ccm import CCM


class CCMRkISP1(CCM):
    hr_name = 'Crosstalk Correction (RkISP1)'
    out_name = 'Ccm'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    # We don't need anything from the config file.
    def validate_config(self, config: dict) -> bool:
        return True

    def process(self, config: dict, images: list, outputs: dict) -> dict:
        output = {}

        ccms = self.do_calibration(images)
        output['ccms'] = ccms

        return output



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
#
# camera tuning tool for AWB

from ctt_image_load import *
import matplotlib.pyplot as plt
from bisect import bisect_left
from scipy.optimize import fmin


"""
obtain piecewise linear approximation for colour curve
"""
def awb(Cam, cal_cr_list, cal_cb_list, plot, grid_size):
    imgs = Cam.imgs
    """
    condense alsc calibration tables into one dictionary
    """
    if cal_cr_list is None:
        colour_cals = None
    else:
        colour_cals = {}
        for cr, cb in zip(cal_cr_list, cal_cb_list):
            cr_tab = cr['table']
            cb_tab = cb['table']
            """
            normalise tables so min value is 1
            """
            cr_tab = cr_tab/np.min(cr_tab)
            cb_tab = cb_tab/np.min(cb_tab)
            colour_cals[cr['ct']] = [cr_tab, cb_tab]
    """
    obtain data from greyscale macbeth patches
    """
    rb_raw = []
    rbs_hat = []
    for Img in imgs:
        Cam.log += '\nProcessing '+Img.name
        """
        get greyscale patches with alsc applied if alsc enabled.
        Note: if alsc is disabled then colour_cals will be set to None and the
        function will just return the greyscale patches
        """
        r_patchs, b_patchs, g_patchs = get_alsc_patches(Img, colour_cals, grid_size=grid_size)
        """
        calculate ratio of r, b to g
        """
        r_g = np.mean(r_patchs/g_patchs)
        b_g = np.mean(b_patchs/g_patchs)
        Cam.log += '\n       r : {:.4f}       b : {:.4f}'.format(r_g, b_g)
        """
        The curve tends to be better behaved in so-called hatspace.
        R, B, G represent the individual channels. The colour curve is plotted in
        r, b space, where:
            r = R/G
            b = B/G
        This will be referred to as dehatspace... (sorry)
        Hatspace is defined as:
            r_hat = R/(R+B+G)
            b_hat = B/(R+B+G)
        To convert from dehatspace to hastpace (hat operation):
            r_hat = r/(1+r+b)
            b_hat = b/(1+r+b)
        To convert from hatspace to dehatspace (dehat operation):
            r = r_hat/(1-r_hat-b_hat)
            b = b_hat/(1-r_hat-b_hat)
        Proof is left as an excercise to the reader...
        Throughout the code, r and b are sometimes referred to as r_g and b_g
        as a reminder that they are ratios
        """
        r_g_hat = r_g/(1+r_g+b_g)
        b_g_hat = b_g/(1+r_g+b_g)
        Cam.log += '\n   r_hat : {:.4f}   b_hat : {:.4f}'.format(r_g_hat, b_g_hat)
        rbs_hat.append((r_g_hat, b_g_hat, Img.col))
        rb_raw.append((r_g, b_g))
        Cam.log += '\n'

    Cam.log += '\nFinished processing images'
    """
    sort all lits simultaneously by r_hat
    """
    rbs_zip = list(zip(rbs_hat, rb_raw))
    rbs_zip.sort(key=lambda x: x[0][0])
    rbs_hat, rb_raw = list(zip(*rbs_zip))
    """
    unzip tuples ready for processing
    """
    rbs_hat = list(zip(*rbs_hat))
    rb_raw = list(zip(*rb_raw))
    """
    fit quadratic fit to r_g hat and b_g_hat
    """
    a, b, c = np.polyfit(rbs_hat[0], rbs_hat[1], 2)
    Cam.log += '\nFit quadratic curve in hatspace'
    """
    the algorithm now approximates the shortest distance from each point to the
    curve in dehatspace. Since the fit is done in hatspace, it is easier to
    find the actual shortest distance in hatspace and use the projection back
    into dehatspace as an overestimate.
    The distance will be used for two things:
        1) In the case that colour temperature does not strictly decrease with
        increasing r/g, the closest point to the line will be chosen out of an
        increasing pair of colours.

        2) To calculate transverse negative an dpositive, the maximum positive
        and negative distance from the line are chosen. This benefits from the
        overestimate as the transverse pos/neg are upper bound values.
    """
    """
    define fit function
    """
    def f(x):
        return a*x**2 + b*x + c
    """
    iterate over points (R, B are x and y coordinates of points) and calculate
    distance to line in dehatspace
    """
    dists = []
    for i, (R, B) in enumerate(zip(rbs_hat[0], rbs_hat[1])):
        """
        define function to minimise as square distance between datapoint and
        point on curve. Squaring is monotonic so minimising radius squared is
        equivalent to minimising radius
        """
        def f_min(x):
            y = f(x)
            return((x-R)**2+(y-B)**2)
        """
        perform optimisation with scipy.optmisie.fmin
        """
        x_hat = fmin(f_min, R, disp=0)[0]
        y_hat = f(x_hat)
        """
        dehat
        """
        x = x_hat/(1-x_hat-y_hat)
        y = y_hat/(1-x_hat-y_hat)
        rr = R/(1-R-B)
        bb = B/(1-R-B)
        """
        calculate euclidean distance in dehatspace
        """
        dist = ((x-rr)**2+(y-bb)**2)**0.5
        """
        return negative if point is below the fit curve
        """
        if (x+y) > (rr+bb):
            dist *= -1
        dists.append(dist)
    Cam.log += '\nFound closest point on fit line to each point in dehatspace'
    """
    calculate wiggle factors in awb. 10% added since this is an upper bound
    """
    transverse_neg = - np.min(dists) * 1.1
    transverse_pos = np.max(dists) * 1.1
    Cam.log += '\nTransverse pos : {:.5f}'.format(transverse_pos)
    Cam.log += '\nTransverse neg : {:.5f}'.format(transverse_neg)
    """
    set minimum transverse wiggles to 0.1 .
    Wiggle factors dictate how far off of the curve the algorithm searches. 0.1
    is a suitable minimum that gives better results for lighting conditions not
    within calibration dataset. Anything less will generalise poorly.
    """
    if transverse_pos < 0.01:
        transverse_pos = 0.01
        Cam.log += '\nForced transverse pos to 0.01'
    if transverse_neg < 0.01:
        transverse_neg = 0.01
        Cam.log += '\nForced transverse neg to 0.01'

    """
    generate new b_hat values at each r_hat according to fit
    """
    r_hat_fit = np.array(rbs_hat[0])
    b_hat_fit = a*r_hat_fit**2 + b*r_hat_fit + c
    """
    transform from hatspace to dehatspace
    """
    r_fit = r_hat_fit/(1-r_hat_fit-b_hat_fit)
    b_fit = b_hat_fit/(1-r_hat_fit-b_hat_fit)
    c_fit = np.round(rbs_hat[2], 0)
    """
    round to 4dp
    """
    r_fit = np.where((1000*r_fit) % 1 <= 0.05, r_fit+0.0001, r_fit)
    r_fit = np.where((1000*r_fit) % 1 >= 0.95, r_fit-0.0001, r_fit)
    b_fit = np.where((1000*b_fit) % 1 <= 0.05, b_fit+0.0001, b_fit)
    b_fit = np.where((1000*b_fit) % 1 >= 0.95, b_fit-0.0001, b_fit)
    r_fit = np.round(r_fit, 4)
    b_fit = np.round(b_fit, 4)
    """
    The following code ensures that colour temperature decreases with
    increasing r/g
    """
    """
    iterate backwards over list for easier indexing
    """
    i = len(c_fit) - 1
    while i > 0:
        if c_fit[i] > c_fit[i-1]:
            Cam.log += '\nColour temperature increase found\n'
            Cam.log += '{} K at r = {} to '.format(c_fit[i-1], r_fit[i-1])
            Cam.log += '{} K at r = {}'.format(c_fit[i], r_fit[i])
            """
            if colour temperature increases then discard point furthest from
            the transformed fit (dehatspace)
            """
            error_1 = abs(dists[i-1])
            error_2 = abs(dists[i])
            Cam.log += '\nDistances from fit:\n'
            Cam.log += '{} K : {:.5f} , '.format(c_fit[i], error_1)
            Cam.log += '{} K : {:.5f}'.format(c_fit[i-1], error_2)
            """
            find bad index
            note that in python false = 0 and true = 1
            """
            bad = i - (error_1 < error_2)
            Cam.log += '\nPoint at {} K deleted as '.format(c_fit[bad])
            Cam.log += 'it is furthest from fit'
            """
            delete bad point
            """
            r_fit = np.delete(r_fit, bad)
            b_fit = np.delete(b_fit, bad)
            c_fit = np.delete(c_fit, bad).astype(np.uint16)
        """
        note that if a point has been discarded then the length has decreased
        by one, meaning that decreasing the index by one will reassess the kept
        point against the next point. It is therefore possible, in theory, for
        two adjacent points to be discarded, although probably rare
        """
        i -= 1

    """
    return formatted ct curve, ordered by increasing colour temperature
    """
    ct_curve = list(np.array(list(zip(b_fit, r_fit, c_fit))).flatten())[::-1]
    Cam.log += '\nFinal CT curve:'
    for i in range(len(ct_curve)//3):
        j = 3*i
        Cam.log += '\n  ct: {}  '.format(ct_curve[j])
        Cam.log += '  r: {}  '.format(ct_curve[j+1])
        Cam.log += '  b: {}  '.format(ct_curve[j+2])

    """
    plotting code for debug
    """
    if plot:
        x = np.linspace(np.min(rbs_hat[0]), np.max(rbs_hat[0]), 100)
        y = a*x**2 + b*x + c
        plt.subplot(2, 1, 1)
        plt.title('hatspace')
        plt.plot(rbs_hat[0], rbs_hat[1], ls='--', color='blue')
        plt.plot(x, y, color='green', ls='-')
        plt.scatter(rbs_hat[0], rbs_hat[1], color='red')
        for i, ct in enumerate(rbs_hat[2]):
            plt.annotate(str(ct), (rbs_hat[0][i], rbs_hat[1][i]))
        plt.xlabel('$\\hat{r}$')
        plt.ylabel('$\\hat{b}$')
        """
        optional set axes equal to shortest distance so line really does
        looks perpendicular and everybody is happy
        """
        # ax = plt.gca()
        # ax.set_aspect('equal')
        plt.grid()
        plt.subplot(2, 1, 2)
        plt.title('dehatspace - indoors?')
        plt.plot(r_fit, b_fit, color='blue')
        plt.scatter(rb_raw[0], rb_raw[1], color='green')
        plt.scatter(r_fit, b_fit, color='red')
        for i, ct in enumerate(c_fit):
            plt.annotate(str(ct), (r_fit[i], b_fit[i]))
        plt.xlabel('$r$')
        plt.ylabel('$b$')
        """
        optional set axes equal to shortest distance so line really does
        looks perpendicular and everybody is happy
        """
        # ax = plt.gca()
        # ax.set_aspect('equal')
        plt.subplots_adjust(hspace=0.5)
        plt.grid()
        plt.show()
    """
    end of plotting code
    """
    return(ct_curve, np.round(transverse_pos, 5), np.round(transverse_neg, 5))


"""
obtain greyscale patches and perform alsc colour correction
"""
def get_alsc_patches(Img, colour_cals, grey=True, grid_size=(16, 12)):
    """
    get patch centre coordinates, image colour and the actual
    patches for each channel, remembering to subtract blacklevel
    If grey then only greyscale patches considered
    """
    grid_w, grid_h = grid_size
    if grey:
        cen_coords = Img.cen_coords[3::4]
        col = Img.col
        patches = [np.array(Img.patches[i]) for i in Img.order]
        r_patchs = patches[0][3::4] - Img.blacklevel_16
        b_patchs = patches[3][3::4] - Img.blacklevel_16
        """
        note two green channels are averages
        """
        g_patchs = (patches[1][3::4]+patches[2][3::4])/2 - Img.blacklevel_16
    else:
        cen_coords = Img.cen_coords
        col = Img.col
        patches = [np.array(Img.patches[i]) for i in Img.order]
        r_patchs = patches[0] - Img.blacklevel_16
        b_patchs = patches[3] - Img.blacklevel_16
        g_patchs = (patches[1]+patches[2])/2 - Img.blacklevel_16

    if colour_cals is None:
        return r_patchs, b_patchs, g_patchs
    """
    find where image colour fits in alsc colour calibration tables
    """
    cts = list(colour_cals.keys())
    pos = bisect_left(cts, col)
    """
    if img colour is below minimum or above maximum alsc calibration colour, simply
    pick extreme closest to img colour
    """
    if pos % len(cts) == 0:
        """
        this works because -0 = 0 = first and -1 = last index
        """
        col_tabs = np.array(colour_cals[cts[-pos//len(cts)]])
        """
    else, perform linear interpolation between existing alsc colour
    calibration tables
    """
    else:
        bef = cts[pos-1]
        aft = cts[pos]
        da = col-bef
        db = aft-col
        bef_tabs = np.array(colour_cals[bef])
        aft_tabs = np.array(colour_cals[aft])
        col_tabs = (bef_tabs*db + aft_tabs*da)/(da+db)
    col_tabs = np.reshape(col_tabs, (2, grid_h, grid_w))
    """
    calculate dx, dy used to calculate alsc table
    """
    w, h = Img.w/2, Img.h/2
    dx, dy = int(-(-(w-1)//grid_w)), int(-(-(h-1)//grid_h))
    """
    make list of pairs of gains for each patch by selecting the correct value
    in alsc colour calibration table
    """
    patch_gains = []
    for cen in cen_coords:
        x, y = cen[0]//dx, cen[1]//dy
        # We could probably do with some better spatial interpolation here?
        col_gains = (col_tabs[0][y][x], col_tabs[1][y][x])
        patch_gains.append(col_gains)

    """
    multiply the r and b channels in each patch by the respective gain, finally
    performing the alsc colour correction
    """
    for i, gains in enumerate(patch_gains):
        r_patchs[i] = r_patchs[i] * gains[0]
        b_patchs[i] = b_patchs[i] * gains[1]

    """
    return greyscale patches, g channel and correct r, b channels
    """
    return r_patchs, b_patchs, g_patchs



#!/usr/bin/env python3
# Copyright 2020 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import json
import os
import os.path
import shutil
import tempfile
import unittest

import check_stable_mojom_compatibility

from mojom.generate import module


class Change:
  """Helper to clearly define a mojom file delta to be analyzed."""

  def __init__(self, filename, old=None, new=None):
    """If old is None, this is a file addition. If new is None, this is a file
    deletion. Otherwise it's a file change."""
    self.filename = filename
    self.old = old
    self.new = new


class UnchangedFile(Change):
  def __init__(self, filename, contents):
    super().__init__(filename, old=contents, new=contents)


class CheckStableMojomCompatibilityTest(unittest.TestCase):
  """Tests covering the behavior of the compatibility checking tool. Note that
  details of different compatibility checks and relevant failure modes are NOT
  covered by these tests. Those are instead covered by unittests in
  version_compatibility_unittest.py. Additionally, the tests which ensure a
  given set of [Stable] mojom definitions are indeed plausibly stable (i.e. they
  have no unstable dependencies) are covered by stable_attribute_unittest.py.

  These tests cover higher-level concerns of the compatibility checking tool,
  like file or symbol, renames, changes spread over multiple files, etc."""

  def verifyBackwardCompatibility(self, changes):
    """Helper for implementing assertBackwardCompatible and
    assertNotBackwardCompatible"""

    temp_dir = tempfile.mkdtemp()
    for change in changes:
      if change.old:
        # Populate the old file on disk in our temporary fake source root
        file_path = os.path.join(temp_dir, change.filename)
        dir_path = os.path.dirname(file_path)
        if not os.path.exists(dir_path):
          os.makedirs(dir_path)
        with open(file_path, 'w') as f:
          f.write(change.old)

    delta = []
    for change in changes:
      if change.old != change.new:
        delta.append({
            'filename': change.filename,
            'old': change.old,
            'new': change.new
        })

    try:
      check_stable_mojom_compatibility.Run(['--src-root', temp_dir],
                                           delta=delta)
    finally:
      shutil.rmtree(temp_dir)

  def assertBackwardCompatible(self, changes):
    self.verifyBackwardCompatibility(changes)

  def assertNotBackwardCompatible(self, changes):
    try:
      self.verifyBackwardCompatibility(changes)
    except Exception:
      return

    raise Exception('Change unexpectedly passed a backward-compatibility check')

  def testBasicCompatibility(self):
    """Minimal smoke test to verify acceptance of a simple valid change."""
    self.assertBackwardCompatible([
        Change('foo/foo.mojom',
               old='[Stable] struct S {};',
               new='[Stable] struct S { [MinVersion=1] int32 x; };')
    ])

  def testBasicIncompatibility(self):
    """Minimal smoke test to verify rejection of a simple invalid change."""
    self.assertNotBackwardCompatible([
        Change('foo/foo.mojom',
               old='[Stable] struct S {};',
               new='[Stable] struct S { int32 x; };')
    ])

  def testIgnoreIfNotStable(self):
    """We don't care about types not marked [Stable]"""
    self.assertBackwardCompatible([
        Change('foo/foo.mojom',
               old='struct S {};',
               new='struct S { int32 x; };')
    ])

  def testRename(self):
    """We can do checks for renamed types."""
    self.assertBackwardCompatible([
        Change('foo/foo.mojom',
               old='[Stable] struct S {};',
               new='[Stable, RenamedFrom="S"] struct T {};')
    ])
    self.assertNotBackwardCompatible([
        Change('foo/foo.mojom',
               old='[Stable] struct S {};',
               new='[Stable, RenamedFrom="S"] struct T { int32 x; };')
    ])
    self.assertBackwardCompatible([
        Change('foo/foo.mojom',
               old='[Stable] struct S {};',
               new="""\
               [Stable, RenamedFrom="S"]
               struct T { [MinVersion=1] int32 x; };
               """)
    ])

  def testNewlyStable(self):
    """We don't care about types newly marked as [Stable]."""
    self.assertBackwardCompatible([
        Change('foo/foo.mojom',
               old='struct S {};',
               new='[Stable] struct S { int32 x; };')
    ])

  def testFileRename(self):
    """Make sure we can still do compatibility checks after a file rename."""
    self.assertBackwardCompatible([
        Change('foo/foo.mojom', old='[Stable] struct S {};', new=None),
        Change('bar/bar.mojom',
               old=None,
               new='[Stable] struct S { [MinVersion=1] int32 x; };')
    ])
    self.assertNotBackwardCompatible([
        Change('foo/foo.mojom', old='[Stable] struct S {};', new=None),
        Change('bar/bar.mojom', old=None, new='[Stable] struct S { int32 x; };')
    ])

  def testWithImport(self):
    """Ensure that cross-module dependencies do not break the compatibility
    checking tool."""
    self.assertBackwardCompatible([
        Change('foo/foo.mojom',
               old="""\
               module foo;
               [Stable] struct S {};
               """,
               new="""\
               module foo;
               [Stable] struct S { [MinVersion=2] int32 x; };
               """),
        Change('bar/bar.mojom',
               old="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; };
               """,
               new="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; [MinVersion=1] int32 y; };
               """)
    ])

  def testWithMovedDefinition(self):
    """If a definition moves from one file to another, we should still be able
    to check compatibility accurately."""
    self.assertBackwardCompatible([
        Change('foo/foo.mojom',
               old="""\
               module foo;
               [Stable] struct S {};
               """,
               new="""\
               module foo;
               """),
        Change('bar/bar.mojom',
               old="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; };
               """,
               new="""\
               module bar;
               import "foo/foo.mojom";
               [Stable, RenamedFrom="foo.S"] struct S {
                 [MinVersion=2] int32 x;
               };
               [Stable] struct T { S s; [MinVersion=1] int32 y; };
               """)
    ])

    self.assertNotBackwardCompatible([
        Change('foo/foo.mojom',
               old="""\
               module foo;
               [Stable] struct S {};
               """,
               new="""\
               module foo;
               """),
        Change('bar/bar.mojom',
               old="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; };
               """,
               new="""\
               module bar;
               import "foo/foo.mojom";
               [Stable, RenamedFrom="foo.S"] struct S { int32 x; };
               [Stable] struct T { S s; [MinVersion=1] int32 y; };
               """)
    ])

  def testWithUnmodifiedImport(self):
    """Unchanged files in the filesystem are still parsed by the compatibility
    checking tool if they're imported by a changed file."""
    self.assertBackwardCompatible([
        UnchangedFile('foo/foo.mojom', 'module foo; [Stable] struct S {};'),
        Change('bar/bar.mojom',
               old="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; };
               """,
               new="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; [MinVersion=1] int32 x; };
               """)
    ])

    self.assertNotBackwardCompatible([
        UnchangedFile('foo/foo.mojom', 'module foo; [Stable] struct S {};'),
        Change('bar/bar.mojom',
               old="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; };
               """,
               new="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; int32 x; };
               """)
    ])

  def testWithPartialImport(self):
    """The compatibility checking tool correctly parses imports with partial
    paths."""
    self.assertBackwardCompatible([
        UnchangedFile('foo/foo.mojom', 'module foo; [Stable] struct S {};'),
        Change('foo/bar.mojom',
               old="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; };
               """,
               new="""\
               module bar;
               import "foo.mojom";
               [Stable] struct T { foo.S s; };
               """)
    ])

    self.assertBackwardCompatible([
        UnchangedFile('foo/foo.mojom', 'module foo; [Stable] struct S {};'),
        Change('foo/bar.mojom',
               old="""\
               module bar;
               import "foo.mojom";
               [Stable] struct T { foo.S s; };
               """,
               new="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; };
               """)
    ])

    self.assertNotBackwardCompatible([
        UnchangedFile('foo/foo.mojom', 'module foo; [Stable] struct S {};'),
        Change('bar/bar.mojom',
               old="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; };
               """,
               new="""\
               module bar;
               import "foo.mojom";
               [Stable] struct T { foo.S s; };
               """)
    ])

    self.assertNotBackwardCompatible([
        UnchangedFile('foo/foo.mojom', 'module foo; [Stable] struct S {};'),
        Change('bar/bar.mojom',
               old="""\
               module bar;
               import "foo.mojom";
               [Stable] struct T { foo.S s; };
               """,
               new="""\
               module bar;
               import "foo/foo.mojom";
               [Stable] struct T { foo.S s; };
               """)
    ])

  def testNewEnumDefault(self):
    # Should be backwards compatible since it does not affect the wire format.
    # This specific case also checks that the backwards compatibility checker
    # does not throw an error due to the older version of the enum not
    # specifying [Default].
    self.assertBackwardCompatible([
        Change('foo/foo.mojom',
               old='[Extensible] enum E { One };',
               new='[Extensible] enum E { [Default] One };')
    ])
    self.assertBackwardCompatible([
        Change('foo/foo.mojom',
               old='[Extensible] enum E { [Default] One, Two, };',
               new='[Extensible] enum E { One, [Default] Two, };')
    ])



# SPDX-License-Identifier: LGPL-2.1-or-later
# Copyright (C) 2022, Tomi Valkeinen <tomi.valkeinen@ideasonboard.com>

import libcamera
from typing import Tuple

class MappedFrameBuffer:
    """
    Provides memoryviews for the FrameBuffer's planes
    """
    def __init__(self, fb: libcamera.FrameBuffer):
        self.__fb = fb
        self.__planes = ()
        self.__maps = ()

    def __enter__(self):
        return self.mmap()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.munmap()

    def mmap(self):
        if self.__planes:
            raise RuntimeError('MappedFrameBuffer already mmapped')

        import os
        import mmap

        fb = self.__fb

        # Collect information about the buffers

        bufinfos = {}

        for plane in fb.planes:
            fd = plane.fd

            if fd not in bufinfos:
                buflen = os.lseek(fd, 0, os.SEEK_END)
                bufinfos[fd] = {'maplen': 0, 'buflen': buflen}
            else:
                buflen = bufinfos[fd]['buflen']

            if plane.offset > buflen or plane.offset + plane.length > buflen:
                raise RuntimeError(f'plane is out of buffer: buffer length={buflen}, ' +
                                   f'plane offset={plane.offset}, plane length={plane.length}')

            bufinfos[fd]['maplen'] = max(bufinfos[fd]['maplen'], plane.offset + plane.length)

        # mmap the buffers

        maps = []

        for fd, info in bufinfos.items():
            map = mmap.mmap(fd, info['maplen'], mmap.MAP_SHARED, mmap.PROT_READ | mmap.PROT_WRITE)
            info['map'] = map
            maps.append(map)

        self.__maps = tuple(maps)

        # Create memoryviews for the planes

        planes = []

        for plane in fb.planes:
            fd = plane.fd
            info = bufinfos[fd]

            mv = memoryview(info['map'])

            start = plane.offset
            end = plane.offset + plane.length

            mv = mv[start:end]

            planes.append(mv)

        self.__planes = tuple(planes)

        return self

    def munmap(self):
        if not self.__planes:
            raise RuntimeError('MappedFrameBuffer not mmapped')

        for p in self.__planes:
            p.release()

        for mm in self.__maps:
            mm.close()

        self.__planes = ()
        self.__maps = ()

    @property
    def planes(self) -> Tuple[memoryview, ...]:
        """memoryviews for the planes"""
        if not self.__planes:
            raise RuntimeError('MappedFrameBuffer not mmapped')

        return self.__planes

    @property
    def fb(self):
        return self.__fb



"""
Some code that will save virtual macbeth charts that show the difference between optimised matrices and non optimised matrices

The function creates an image that is 1550 by 1050 pixels wide, and fills it with patches which are 200x200 pixels in size
Each patch contains the ideal color, the color from the original matrix, and the color from the final matrix
_________________
|               |
|  Ideal Color  |
|_______________|
|  Old  |  new  |
| Color | Color |
|_______|_______|

Nice way of showing how the optimisation helps change the colors and the color matricies
"""
import numpy as np
from PIL import Image


def visualise_macbeth_chart(macbeth_rgb, original_rgb, new_rgb, output_filename):
    image = np.zeros((1050, 1550, 3), dtype=np.uint8)
    colorindex = -1
    for y in range(6):
        for x in range(4):  # Creates 6 x 4 grid of macbeth chart
            colorindex += 1
            xlocation = 50 + 250 * x  # Means there is 50px of black gap between each square, more like the real macbeth chart.
            ylocation = 50 + 250 * y
            for g in range(200):
                for i in range(100):
                    image[xlocation + i, ylocation + g] = macbeth_rgb[colorindex]
            xlocation = 150 + 250 * x
            ylocation = 50 + 250 * y
            for i in range(100):
                for g in range(100):
                    image[xlocation + i, ylocation + g] = original_rgb[colorindex]  # Smaller squares below to compare the old colors with the new ones
            xlocation = 150 + 250 * x
            ylocation = 150 + 250 * y
            for i in range(100):
                for g in range(100):
                    image[xlocation + i, ylocation + g] = new_rgb[colorindex]

    img = Image.fromarray(image, 'RGB')
    img.save(str(output_filename) + 'Generated Macbeth Chart.png')



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2019, Google Inc.
#
# Author: Laurent Pinchart <laurent.pinchart@ideasonboard.com>
#
# Helper classes to handle source code generation for libcamera controls


class ControlEnum(object):
    def __init__(self, data):
        self.__data = data

    @property
    def description(self):
        """The enum description"""
        return self.__data.get('description')

    @property
    def name(self):
        """The enum name"""
        return self.__data.get('name')

    @property
    def value(self):
        """The enum value"""
        return self.__data.get('value')


class Control(object):
    def __init__(self, name, data, vendor, mode):
        self.__name = name
        self.__data = data
        self.__enum_values = None
        self.__size = None
        self.__vendor = vendor

        enum_values = data.get('enum')
        if enum_values is not None:
            self.__enum_values = [ControlEnum(enum) for enum in enum_values]

        size = self.__data.get('size')
        if size is not None:
            if len(size) == 0:
                raise RuntimeError(f'Control `{self.__name}` size must have at least one dimension')

            # Compute the total number of elements in the array. If any of the
            # array dimension is a string, the array is variable-sized.
            num_elems = 1
            for dim in size:
                if type(dim) is str:
                    num_elems = 0
                    break

                dim = int(dim)
                if dim <= 0:
                    raise RuntimeError(f'Control `{self.__name}` size must have positive values only')

                num_elems *= dim

            self.__size = num_elems

        if mode == 'properties':
            self.__direction = 'out'
        else:
            direction = self.__data.get('direction')
            if direction is None:
                raise RuntimeError(f'Control `{self.__name}` missing required field `{direction}`')
            if direction not in ['in', 'out', 'inout']:
                raise RuntimeError(f'Control `{self.__name}` direction `{direction}` is invalid; must be one of `in`, `out`, or `inout`')
            self.__direction = direction

    @property
    def description(self):
        """The control description"""
        return self.__data.get('description')

    @property
    def enum_values(self):
        """The enum values, if the control is an enumeration"""
        if self.__enum_values is None:
            return
        for enum in self.__enum_values:
            yield enum

    @property
    def enum_values_count(self):
        """The number of enum values, if the control is an enumeration"""
        if self.__enum_values is None:
            return 0
        return len(self.__enum_values)

    @property
    def is_enum(self):
        """Is the control an enumeration"""
        return self.__enum_values is not None

    @property
    def vendor(self):
        """The vendor string, or None"""
        return self.__vendor

    @property
    def name(self):
        """The control name (CamelCase)"""
        return self.__name

    @property
    def type(self):
        typ = self.__data.get('type')
        size = self.__data.get('size')

        if typ == 'string':
            return 'std::string'

        if self.__size is None:
            return typ

        if self.__size:
            return f"Span<const {typ}, {self.__size}>"
        else:
            return f"Span<const {typ}>"

    @property
    def direction(self):
        in_flag = 'ControlId::Direction::In'
        out_flag = 'ControlId::Direction::Out'

        if self.__direction == 'inout':
            return f'{in_flag} | {out_flag}'
        if self.__direction == 'in':
            return in_flag
        if self.__direction == 'out':
            return out_flag

    @property
    def element_type(self):
        return self.__data.get('type')

    @property
    def size(self):
        return self.__size
